import functools as fct
import warnings

import cv2
import cvxpy as cvx
import dask as da
import dask.array as darr
import networkx as nx
import numpy as np
import xarray as xr
from dask import delayed
from scipy.linalg import lstsq, toeplitz
from scipy.ndimage import label
from scipy.signal import butter, lfilter, welch
from scipy.sparse import dia_matrix, diags
from skimage import morphology as moph
from sklearn.linear_model import LassoLars
from sklearn.utils import parallel_backend
from statsmodels.tsa.stattools import acovf
from .utilities import rechunk_like


def get_noise_fft(varr, noise_range=(0.25, 0.5), noise_method="logmexp"):
    """[summary]

    Args:
        varr ([type]): [description]
        noise_range (tuple, optional): [description]. Defaults to (0.25, 0.5).
        noise_method (str, optional): [description]. Defaults to 'logmexp'.

    Returns:
        [type]: [description]
    """
    sn = xr.apply_ufunc(
        _noise_fft,
        varr.chunk(dict(frame=-1)),
        input_core_dims=[["frame"]],
        output_core_dims=[[]],
        dask="parallelized",
        vectorize=True,
        kwargs=dict(noise_range=noise_range, noise_method=noise_method),
        output_dtypes=[np.float],
    )
    return sn


def _noise_fft(px, noise_range=(0.25, 0.5), noise_method="logmexp"):
    """[summary]

    Args:
        px ([type]): [description]
        noise_range (tuple, optional): [description]. Defaults to (0.25, 0.5).
        noise_method (str, optional): [description]. Defaults to 'logmexp'.

    Returns:
        [type]: [description]
    """
    _T = len(px)
    nr = np.around(np.array(noise_range) * 2 * _T).astype(int)
    px_fft = np.fft.rfft(px)
    px_psd = 1 / _T * np.abs(px_fft) ** 2
    px_band = px_psd[nr[0] : nr[1]]
    if noise_method == "mean":
        return np.sqrt(px_band.mean())
    elif noise_method == "median":
        return np.sqrt(px_band.median())
    elif noise_method == "logmexp":
        eps = np.finfo(px_band.dtype).eps
        return np.sqrt(np.exp(np.log(px_band + eps).mean()))
    elif noise_method == "sum":
        return np.sqrt(px_band.sum())


def get_noise_welch(
    varr, noise_range=(0.25, 0.5), noise_method="logmexp", compute=True
):
    print("estimating noise")
    sn = xr.apply_ufunc(
        noise_welch,
        varr.chunk(dict(frame=-1)),
        input_core_dims=[["frame"]],
        dask="parallelized",
        vectorize=True,
        kwargs=dict(noise_range=noise_range, noise_method=noise_method),
        output_dtypes=[varr.dtype],
    )
    if compute:
        sn = sn.compute()
    return sn


def noise_welch(y, noise_range, noise_method):
    """[summary]

    Args:
        y ([type]): [description]
        noise_range ([type]): [description]
        noise_method ([type]): [description]

    Returns:
        [type]: [description]
    """
    ff, Pxx = welch(y)
    mask0, mask1 = ff > noise_range[0], ff < noise_range[1]
    mask = np.logical_and(mask0, mask1)
    Pxx_ind = Pxx[mask]
    sn = {
        "mean": lambda x: np.sqrt(np.mean(x / 2)),
        "median": lambda x: np.sqrt(np.median(x / 2)),
        "logmexp": lambda x: np.sqrt(np.exp(np.mean(np.log(x / 2)))),
    }[noise_method](Pxx_ind)
    return sn


def update_spatial(
    Y,
    A,
    b,
    C,
    f,
    sn,
    gs_sigma=6,
    dl_wnd=5,
    sparse_penal=0.5,
    update_background=True,
    post_scal=False,
    normalize=True,
    zero_thres="eps",
    sched="single-threaded",
):
    """[summary]

    Args:
        Y ([type]): [description]
        A ([type]): [description]
        b ([type]): [description]
        C ([type]): [description]
        f ([type]): [description]
        sn ([type]): [description]
        gs_sigma (int, optional): [description]. Defaults to 6.
        dl_wnd (int, optional): [description]. Defaults to 5.
        sparse_penal (float, optional): [description]. Defaults to 0.5.
        update_background (bool, optional): [description]. Defaults to True.
        post_scal (bool, optional): [description]. Defaults to False.
        normalize (bool, optional): [description]. Defaults to True.
        zero_thres (str, optional): [description]. Defaults to 'eps'.
        sched (str, optional): [description]. Defaults to 'single-threaded'.

    Returns:
        [type]: [description]
    """
    _T = len(Y.coords["frame"])
    print("estimating penalty parameter")
    cct = C.dot(C, "frame")
    alpha = sparse_penal * sn * np.sqrt(np.max(np.diag(cct))) / _T
    alpha = alpha.persist()
    print("computing subsetting matrix")
    if dl_wnd:
        selem = moph.disk(dl_wnd)
        sub = xr.apply_ufunc(
            cv2.dilate,
            A.chunk(dict(height=-1, width=-1)),
            input_core_dims=[["height", "width"]],
            output_core_dims=[["height", "width"]],
            vectorize=True,
            kwargs=dict(kernel=selem),
            dask="parallelized",
            output_dtypes=[A.dtype],
        )
        sub = sub > 0
    else:
        sub = xr.apply_ufunc(np.ones_like, A.compute())
    if update_background:
        A = xr.concat([A, b.assign_coords(unit_id=-1)], "unit_id")
        b_erd = xr.apply_ufunc(
            cv2.erode,
            b.chunk(dict(height=-1, width=-1)),
            input_core_dims=[["height", "width"]],
            output_core_dims=[["height", "width"]],
            kwargs=dict(kernel=selem),
            dask="parallelized",
            output_dtypes=[b.dtype],
        )
        sub = xr.concat(
            [sub, (b_erd > 0).astype(bool).assign_coords(unit_id=-1)], "unit_id"
        )
        C = xr.concat([C, f.assign_coords(unit_id=-1)], "unit_id")
    sub = sub.persist()
    print("fitting spatial matrix")
    A_new = xr.apply_ufunc(
        update_spatial_perpx,
        Y.chunk(dict(frame=-1)),
        alpha,
        sub.chunk(dict(unit_id=-1)),
        C.chunk(dict(frame=-1, unit_id=-1)),
        input_core_dims=[["frame"], [], ["unit_id"], ["frame", "unit_id"]],
        output_core_dims=[["unit_id"]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[Y.dtype],
    )
    try:
        with parallel_backend("dask"):
            A_new = A_new.persist()
    except ValueError:
        with da.config.set(scheduler=sched):
            A_new = A_new.persist()
    print("removing empty units")
    if zero_thres == "eps":
        zero_thres = np.finfo(A_new.dtype).eps
    A_new = A_new.where(A_new > zero_thres).fillna(0)
    non_empty = (A_new.sum(["width", "height"]) > 0).compute()
    A_new = A_new.where(non_empty, drop=True)
    C_new = C.where(non_empty, drop=True)
    A_new = rechunk_like(A_new, A).persist()
    C_new = rechunk_like(C_new, C).persist()
    if post_scal and len(A_new) > 0:
        print("post-hoc scaling")
        A_new_flt = A_new.stack(spatial=["height", "width"]).compute()
        Y_flt = Y.mean("frame").stack(spatial=["height", "width"]).compute()

        def lstsq(a, b):
            return np.linalg.lstsq(a, b, rcond=-1)[0]

        scale = xr.apply_ufunc(
            lstsq,
            A_new_flt,
            Y_flt,
            input_core_dims=[["spatial", "unit_id"], ["spatial"]],
            output_core_dims=[["unit_id"]],
        )
        C_mean = C.mean("frame").compute()
        scale = scale / C_mean
        A_new = A_new * scale
        try:
            A_new = A_new.persist()
        except np.linalg.LinAlgError:
            warnings.warn("post-hoc scaling failed", RuntimeWarning)
    if update_background:
        print("updating background")
        try:
            b_new = A_new.sel(unit_id=-1)
            b_new = b_new / da.array.linalg.norm(b_new.data)
            f_new = xr.apply_ufunc(
                da.array.tensordot,
                Y,
                b_new,
                input_core_dims=[["frame", "height", "width"], ["height", "width"]],
                output_core_dims=[["frame"]],
                kwargs=dict(axes=[(1, 2), (0, 1)]),
                dask="allowed",
            ).persist()
            A_new = A_new.drop(-1, "unit_id")
            C_new = C_new.drop(-1, "unit_id")
        except KeyError:
            print("background terms are empty")
            b_new = xr.zeros_like(b)
            f_new = xr.zeros_like(f)
    else:
        b_new = b
        f_new = f
    if normalize and len(A_new) > 0:
        print("normalizing result")
        A_norm = xr.apply_ufunc(
            darr.linalg.norm,
            A_new.stack(spatial=["height", "width"]),
            input_core_dims=[["spatial", "unit_id"]],
            output_core_dims=[["unit_id"]],
            kwargs=dict(axis=0),
            dask="allowed",
        )
        A_new = (A_new / A_norm).persist()
    return A_new, b_new, C_new, f_new


def update_spatial_perpx(y, alpha, sub, C):
    """[summary]

    Args:
        y ([type]): [description]
        alpha ([type]): [description]
        sub ([type]): [description]
        C ([type]): [description]

    Returns:
        [type]: [description]
    """
    res = np.zeros_like(sub, dtype=y.dtype)
    if np.sum(sub) > 0:
        C = C[:, sub]
        clf = LassoLars(alpha=alpha, positive=True)
        coef = clf.fit(C, y).coef_
        res[np.where(sub)[0]] = coef
    return res


def compute_trace(Y, A, b, C, f, noise_freq=None):
    """[summary]

    Args:
        Y ([type]): [description]
        A ([type]): [description]
        b ([type]): [description]
        C ([type]): [description]
        f ([type]): [description]
        noise_freq ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    nunits = len(A.coords["unit_id"])
    A_rechk = A.chunk(dict(height=-1, width=-1))
    C_rechk = C.chunk(dict(unit_id=-1))
    Y_rechk = Y.chunk(dict(height=-1, width=-1))
    AA = xr.apply_ufunc(
        da.array.tensordot,
        A_rechk,
        A_rechk.rename(dict(unit_id="unit_id_cp")),
        input_core_dims=[
            ["unit_id", "height", "width"],
            ["height", "width", "unit_id_cp"],
        ],
        output_core_dims=[["unit_id", "unit_id_cp"]],
        dask="allowed",
        kwargs=dict(axes=([1, 2], [0, 1])),
        output_dtypes=[A.dtype],
    )
    nA = (A_rechk ** 2).sum(["height", "width"]).compute()
    nA_inv = xr.apply_ufunc(
        lambda x: np.asarray(diags(x).todense()),
        1 / nA,
        input_core_dims=[["unit_id"]],
        output_core_dims=[["unit_id", "unit_id_cp"]],
        dask="parallelized",
        output_dtypes=[nA.dtype],
        output_sizes=dict(unit_id_cp=nunits),
    ).compute()
    nA_inv = nA_inv.assign_coords(unit_id_temp=AA.coords["unit_id_cp"])
    b = b.fillna(0).expand_dims("dot").chunk(dict(height=-1, width=-1))
    f = f.fillna(0).expand_dims("dot")
    B = xr.apply_ufunc(
        da.array.dot,
        b,
        f,
        input_core_dims=[["height", "width", "dot"], ["dot", "frame"]],
        output_core_dims=[["height", "width", "frame"]],
        dask="allowed",
        output_dtypes=[b.dtype],
    )
    Y = Y_rechk - B
    YA = xr.apply_ufunc(
        da.array.tensordot,
        Y,
        A_rechk,
        input_core_dims=[["frame", "height", "width"], ["height", "width", "unit_id"]],
        output_core_dims=[["frame", "unit_id"]],
        dask="allowed",
        kwargs=dict(axes=([1, 2], [0, 1])),
        output_dtypes=[A.dtype],
    ).rename(dict(unit_id="unit_id_cp"))
    YA_norm = xr.apply_ufunc(
        da.array.dot,
        YA,
        nA_inv,
        input_core_dims=[["frame", "unit_id_cp"], ["unit_id_cp", "unit_id"]],
        output_core_dims=[["frame", "unit_id"]],
        dask="allowed",
        output_dtypes=[YA.dtype],
    )
    CA = xr.apply_ufunc(
        da.array.dot,
        C_rechk,
        AA.chunk(dict(unit_id=-1, unit_id_cp=-1)),
        input_core_dims=[["frame", "unit_id"], ["unit_id", "unit_id_cp"]],
        output_core_dims=[["frame", "unit_id_cp"]],
        dask="allowed",
        output_dtypes=[C.dtype],
    )
    CA_norm = xr.apply_ufunc(
        da.array.dot,
        CA,
        nA_inv,
        input_core_dims=[["frame", "unit_id_cp"], ["unit_id_cp", "unit_id"]],
        output_core_dims=[["frame", "unit_id"]],
        dask="allowed",
        output_dtypes=[CA.dtype],
    )
    YrA = YA_norm - CA_norm + C_rechk
    if noise_freq:
        print("smoothing signals")
        but_b, but_a = butter(2, noise_freq, btype="low", analog=False)
        YrA_smth = xr.apply_ufunc(
            lambda x: lfilter(but_b, but_a, x),
            YrA.chunk(dict(frame=-1)),
            input_core_dims=[["frame"]],
            output_core_dims=[["frame"]],
            vectorize=True,
            dask="parallelized",
            output_dtypes=[YrA.dtype],
        )
    else:
        YrA_smth = YrA
    return YrA_smth


def update_temporal(
    Y,
    A,
    b,
    C,
    f,
    sn_spatial,
    YrA=None,
    noise_freq=0.25,
    p=None,
    add_lag="p",
    jac_thres=0.1,
    use_spatial=False,
    sparse_penal=1,
    bseg=None,
    zero_thres=1e-8,
    max_iters=200,
    use_smooth=True,
    compute=True,
    normalize=True,
    post_scal=True,
    scs_fallback=False,
    sched="processes",
):
    """[summary]

    Args:
        Y ([type]): [description]
        A ([type]): [description]
        b ([type]): [description]
        C ([type]): [description]
        f ([type]): [description]
        sn_spatial ([type]): [description]
        YrA ([type], optional): [description]. Defaults to None.
        noise_freq (float, optional): [description]. Defaults to 0.25.
        p ([type], optional): [description]. Defaults to None.
        add_lag (str, optional): [description]. Defaults to 'p'.
        jac_thres (float, optional): [description]. Defaults to 0.1.
        use_spatial (bool, optional): [description]. Defaults to False.
        sparse_penal (int, optional): [description]. Defaults to 1.
        zero_thres ([type], optional): [description]. Defaults to 1e-8.
        max_iters (int, optional): [description]. Defaults to 200.
        use_smooth (bool, optional): [description]. Defaults to True.
        compute (bool, optional): [description]. Defaults to True.
        normalize (bool, optional): [description]. Defaults to True.
        post_scal (bool, optional): [description]. Defaults to True.
        scs_fallback (bool, optional): [description]. Defaults to False.
        sched (str, optional): [description]. Defaults to 'processes'.

    Returns:
        [type]: [description]
    """
    print("grouping overlaping units")
    A_pos = (A > 0).astype(float)
    A_neg = (A == 0).astype(float)
    A_inter = xr.apply_ufunc(
        da.array.tensordot,
        A_pos,
        A_pos.rename(unit_id="unit_id_cp"),
        input_core_dims=[
            ["unit_id", "height", "width"],
            ["height", "width", "unit_id_cp"],
        ],
        output_core_dims=[["unit_id", "unit_id_cp"]],
        dask="allowed",
        kwargs=dict(axes=([1, 2], [0, 1])),
        output_dtypes=[A_pos.dtype],
    )
    A_union = xr.apply_ufunc(
        da.array.tensordot,
        A_neg,
        A_neg.rename(unit_id="unit_id_cp"),
        input_core_dims=[
            ["unit_id", "height", "width"],
            ["height", "width", "unit_id_cp"],
        ],
        output_core_dims=[["unit_id", "unit_id_cp"]],
        dask="allowed",
        kwargs=dict(axes=([1, 2], [0, 1])),
        output_dtypes=[A_neg.dtype],
    )
    A_jac = A_inter / (A.sizes["height"] * A.sizes["width"] - A_union)
    if compute:
        A_jac = A_jac.compute()
    unit_labels = xr.apply_ufunc(
        label_connected,
        A_jac > jac_thres,
        input_core_dims=[["unit_id", "unit_id_cp"]],
        kwargs=dict(only_connected=True),
        output_core_dims=[["unit_id"]],
    )
    if YrA is not None:
        YrA = YrA
    else:
        print("computing trace")
        YrA = compute_trace(Y, A, b, C, f).persist()
    YrA = YrA.chunk(dict(frame=-1, unit_id=1))
    YrA = YrA.assign_coords(unit_labels=unit_labels)
    if normalize:
        print("normalizing traces")
        YrA_norm = (YrA / YrA.sum("frame") * YrA.sizes["frame"]).persist()
    else:
        YrA_norm = YrA
    sn_temp = get_noise_fft(
        YrA_norm, noise_range=(noise_freq, 1), noise_method="sum"
    ).persist()
    sn_temp = sn_temp.assign_coords(unit_labels=unit_labels)
    if use_spatial:
        print("flattening spatial dimensions")
        Y_flt = Y.stack(spatial=("height", "width"))
        A_flt = A.stack(spatial=("height", "width")).assign_coords(
            unit_labels=unit_labels
        )
        sn_spatial = sn_spatial.stack(spatial=("height", "width"))
    if use_smooth:
        print("smoothing signals")
        but_b, but_a = butter(2, noise_freq, btype="low", analog=False)
        YrA_smth = xr.apply_ufunc(
            lambda x: lfilter(but_b, but_a, x),
            YrA_norm,
            input_core_dims=[["frame"]],
            output_core_dims=[["frame"]],
            vectorize=True,
            dask="parallelized",
            output_dtypes=[YrA.dtype],
        )
        if compute:
            YrA_smth = YrA_smth.persist()
            sn_temp_smth = get_noise_fft(
                YrA_smth, noise_range=(noise_freq, 1)
            ).persist()
            sn_temp_smth = sn_temp_smth.assign_coords(unit_labels=unit_labels)
    else:
        YrA_smth = YrA_norm
        sn_temp_smth = sn_temp
    if p is None:
        print("estimating order p for each neuron")
        p = xr.apply_ufunc(
            get_p,
            YrA_smth,
            input_core_dims=[["frame"]],
            vectorize=True,
            dask="parallelized",
            output_dtypes=[np.int],
        ).clip(1)
        if compute:
            p = p.compute()
        p_max = p.max().values
    else:
        p_max = p
    print("estimating AR coefficients")
    g = xr.apply_ufunc(
        get_ar_coef,
        YrA_smth.chunk(dict(frame=-1)),
        sn_temp_smth,
        p,
        input_core_dims=[["frame"], [], []],
        output_core_dims=[["lag"]],
        kwargs=dict(pad=p_max, add_lag=add_lag),
        vectorize=True,
        dask="parallelized",
        output_dtypes=[sn_temp_smth.dtype],
        output_sizes=dict(lag=p_max),
    )
    g = g.assign_coords(lag=np.arange(1, p_max + 1), unit_labels=unit_labels)
    if compute:
        g = g.persist()
    print("updating isolated temporal components")
    if use_spatial == "full":
        result_iso = xr.apply_ufunc(
            update_temporal_cvxpy,
            Y_flt.chunk(dict(spatial=-1, frame=-1)),
            g.where(unit_labels == -1, drop=True).chunk(dict(lag=-1)),
            sn_spatial.chunk(dict(spatial=-1)),
            A_flt.where(unit_labels == -1, drop=True).chunk(dict(spatial=-1)),
            input_core_dims=[["spatial", "frame"], ["lag"], ["spatial"], ["spatial"]],
            output_core_dims=[["trace", "frame"]],
            vectorize=True,
            dask="parallelized",
            kwargs=dict(
                sparse_penal=sparse_penal,
                max_iters=max_iters,
                scs_fallback=scs_fallback,
            ),
            output_sizes=dict(trace=5),
            output_dtypes=[YrA.dtype],
        )
    else:
        gu_update = darr.gufunc(
            fct.partial(
                update_temporal_cvxpy,
                sparse_penal=sparse_penal,
                max_iters=max_iters,
                bseg=bseg,
                scs_fallback=scs_fallback,
            ),
            signature="(f),(l),()->(t,f)",
            vectorize=True,
            output_dtypes=[YrA.dtype],
            output_sizes=dict(t=5),
        )
        result_iso = xr.apply_ufunc(
            gu_update,
            YrA_norm.where(unit_labels == -1, drop=True).persist(),
            g.where(unit_labels == -1, drop=True).chunk(dict(lag=-1)).persist(),
            sn_temp.where(unit_labels == -1, drop=True).persist(),
            input_core_dims=[["frame"], ["lag"], []],
            output_core_dims=[["trace", "frame"]],
            dask="allowed",
        )
    if compute:
        with da.config.set(scheduler=sched):
            result_iso = result_iso.compute()
    print("updating overlapping temporal components")
    res_list = []
    g_ovlp = g.where(unit_labels >= 0, drop=True)
    if len(g_ovlp) > 0:
        for cur_labl, cur_g in g_ovlp.groupby("unit_labels"):
            if use_spatial:
                cur_A = A_flt_ovlp.where(unit_labels == cur_labl, drop=True)
                cur_res = delayed(xr.apply_ufunc)(
                    update_temporal_cvxpy,
                    Y_flt.chunk(dict(spatial=-1, frame=-1)),
                    cur_g.chunk(dict(lag=-1)),
                    sn_spatial.chunk(dict(spatial=-1)),
                    cur_A.chunk(dict(spatial=-1)),
                    input_core_dims=[
                        ["spatial", "frame"],
                        ["unit_id", "lag"],
                        ["spatial"],
                        ["unit_id", "spatial"],
                    ],
                    output_core_dims=[["trace", "unit_id", "frame"]],
                    dask="parallelized",
                    kwargs=dict(
                        sparse_penal=sparse_penal,
                        max_iters=max_iters,
                        scs_fallback=scs_fallback,
                    ),
                    output_sizes=dict(trace=5),
                    output_dtypes=[YrA.dtype],
                )
            else:
                cur_YrA = YrA_norm.where(unit_labels == cur_labl, drop=True)
                cur_sn_temp = sn_temp.where(unit_labels == cur_labl, drop=True)
                cur_res = delayed(xr.apply_ufunc)(
                    update_temporal_cvxpy,
                    cur_YrA.compute(),
                    cur_g.compute(),
                    cur_sn_temp.compute(),
                    input_core_dims=[
                        ["unit_id", "frame"],
                        ["unit_id", "lag"],
                        ["unit_id"],
                    ],
                    output_core_dims=[["trace", "unit_id", "frame"]],
                    dask="forbidden",
                    kwargs=dict(
                        sparse_penal=sparse_penal,
                        max_iters=max_iters,
                        bseg=bseg,
                        scs_fallback=scs_fallback,
                    ),
                    output_sizes=dict(trace=5),
                    output_dtypes=[YrA.dtype],
                )
                res_list.append(cur_res)
        if compute:
            with da.config.set(scheduler=sched):
                (result_ovlp,) = da.compute(res_list)
                result = (
                    xr.concat(result_ovlp + [result_iso], "unit_id")
                    .sortby("unit_id")
                    .drop("unit_labels")
                )
    else:
        result = result_iso.sortby("unit_id").drop("unit_labels")
    C_new = result.isel(trace=0).dropna("unit_id")
    S_new = result.isel(trace=1).dropna("unit_id")
    B_new = result.isel(trace=2).dropna("unit_id").squeeze()
    C0_new = result.isel(trace=3, frame=0).dropna("unit_id").squeeze()
    dc_new = result.isel(trace=4).dropna("unit_id")
    g_new = g.sel(unit_id=C_new.coords["unit_id"]).drop("unit_labels")
    if zero_thres:
        mask = S_new.where(S_new > zero_thres).fillna(0).sum("frame").astype(bool)
        mask_coord = mask.where(~mask, drop=True).coords["unit_id"].values
        print(
            "{} units dropped due to poor fit:\n {}".format(
                len(mask_coord), str(mask_coord)
            )
        )
    else:
        mask_coord = S_new.coords["unit_id"].values
        mask = xr.DataArray(
            np.ones(len(mask_coord)), dims=["unit_id"], coords=dict(unit_id=mask_coord)
        )
    C_new, S_new, C0_new, B_new, dc_new = (
        C_new.where(mask, drop=True),
        S_new.where(mask, drop=True),
        C0_new.where(mask, drop=True),
        B_new.where(mask, drop=True),
        dc_new.where(mask, drop=True),
    )
    YrA_new = YrA.drop("unit_labels").sel(unit_id=C_new.coords["unit_id"])
    sig_new = (C0_new * dc_new + B_new + C_new).persist()
    if post_scal and len(sig_new) > 0:
        print("post-hoc scaling")

        def lstsq(a, b):
            a = np.atleast_2d(a).T
            return np.linalg.lstsq(a, b, rcond=-1)[0]

        scal = xr.apply_ufunc(
            lstsq,
            sig_new.chunk(dict(frame=-1)),
            YrA_new.chunk(dict(frame=-1)),
            input_core_dims=[["frame"], ["frame"]],
            output_core_dims=[[]],
            vectorize=True,
            dask="parallelized",
            output_dtypes=[C_new.dtype],
        )
        scal = scal.persist()
        C_new = (C_new * scal).persist()
        S_new = (S_new * scal).persist()
        B_new = (B_new * scal).persist()
        C0_new = (C0_new * scal).persist()
        sig_new = (sig_new * scal).persist()
    else:
        scal = None
    if len(sig_new) > 0:
        C_new = rechunk_like(C_new.persist(), C)
        S_new = rechunk_like(S_new.persist(), C)
        B_new = rechunk_like(B_new.persist(), C)
        C0_new = rechunk_like(C0_new.persist(), C)
        g_new = rechunk_like(g_new.persist(), C)
        sig_new = rechunk_like(sig_new.persist(), C)
    return (YrA_norm, C_new, S_new, B_new, C0_new, sig_new, g_new, scal)


def get_ar_coef(y, sn, p, add_lag, pad=None):
    """[summary]

    Args:
        y ([type]): [description]
        sn ([type]): [description]
        p ([type]): [description]
        add_lag ([type]): [description]
        pad ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    if add_lag == "p":
        max_lag = p * 2
    else:
        max_lag = p + add_lag
    cov = acovf(y, fft=True)
    C_mat = toeplitz(cov[:max_lag], cov[:p]) - sn ** 2 * np.eye(max_lag, p)
    g = lstsq(C_mat, cov[1 : max_lag + 1])[0]
    if pad:
        res = np.zeros(pad)
        res[: len(g)] = g
        return res
    else:
        return g


def get_p(y):
    """[summary]

    Args:
        y ([type]): [description]

    Returns:
        [type]: [description]
    """
    dif = np.append(np.diff(y), 0)
    rising = dif > 0
    prd_ris, num_ris = label(rising)
    ext_prd = np.zeros(num_ris)
    for id_prd in range(num_ris):
        prd = y[prd_ris == id_prd + 1]
        ext_prd[id_prd] = prd[-1] - prd[0]
    id_max_prd = np.argmax(ext_prd)
    return np.sum(rising[prd_ris == id_max_prd + 1])


def update_temporal_cvxpy(y, g, sn, A=None, bseg=None, **kwargs):
    """[summary]

    spatial:
    (d, f), (u, p), (d), (d, u)
    (d, f), (p), (d), (d)
    trace:
    (u, f), (u, p), (u)
    (f), (p), ()

    Args:
        y ([type]): [description]
        g ([type]): [description]
        sn ([type]): [description]
        A ([type], optional): [description]. Defaults to None.
        bseg ([type], optional): [description]. Defaults to None

    Raises:
        ValueError: [description]
        ValueError: [description]
        ValueError: [description]

    Returns:
        [type]: [description]
    """
    # get_parameters
    sparse_penal = kwargs.get("sparse_penal")
    max_iters = kwargs.get("max_iters")
    use_cons = kwargs.get("use_cons", False)
    scs = kwargs.get("scs_fallback")
    # conform variables to generalize multiple unit case
    if y.ndim < 2:
        y = y.reshape((1, -1))
    if g.ndim < 2:
        g = g.reshape((1, -1))
    sn = np.atleast_1d(sn)
    if A is not None:
        if A.ndim < 2:
            A = A.reshape((-1, 1))
    # get count of frames and units
    _T = y.shape[-1]
    _u = g.shape[0]
    if A is not None:
        _d = A.shape[0]
    # construct G matrix and decay vector per unit
    dc_vec = np.zeros((_u, _T))
    G_ls = []
    for cur_u in range(_u):
        cur_g = g[cur_u, :]
        # construct first column and row
        cur_c, cur_r = np.zeros(_T), np.zeros(_T)
        cur_c[0] = 1
        cur_r[0] = 1
        cur_c[1 : len(cur_g) + 1] = -cur_g
        # update G with toeplitz matrix
        G_ls.append(cvx.Constant(dia_matrix(toeplitz(cur_c, cur_r))))
        # update dc_vec
        cur_gr = np.roots(cur_c)
        dc_vec[cur_u, :] = np.max(cur_gr.real) ** np.arange(_T)
    # get noise threshold
    thres_sn = sn * np.sqrt(_T)
    # construct variables
    if bseg is not None:
        nseg = int(np.max(bseg) + 1)
        b_temp = np.zeros((nseg, _T))
        for iseg in range(nseg):
            b_temp[iseg, bseg == iseg] = 1
        b_cmp = cvx.Variable((_u, nseg))
    else:
        b_temp = np.ones((1, _T))
        b_cmp = cvx.Variable((_u, 1))
    b = b_cmp @ b_temp  # baseline fluorescence per unit
    c0 = cvx.Variable(_u)  # initial fluorescence per unit
    c = cvx.Variable((_u, _T))  # calcium trace per unit
    s = cvx.vstack([G_ls[u] * c[u, :] for u in range(_u)])  # spike train per unit
    # residual noise per unit
    if A is not None:
        sig = cvx.vstack(
            [
                (A * c)[px, :] + (A * b)[px, :] + (A * cvx.diag(c0) * dc_vec)[px, :]
                for px in range(_d)
            ]
        )
        noise = y - sig
    else:
        sig = cvx.vstack([c[u, :] + b[u, :] + c0[u] * dc_vec[u, :] for u in range(_u)])
        noise = y - sig
    noise = cvx.vstack([cvx.norm(noise[i, :], 2) for i in range(noise.shape[0])])
    # construct constraints
    cons = []
    cons.append(
        b >= np.broadcast_to(np.min(y, axis=-1).reshape((-1, 1)), y.shape)
    )  # baseline larger than minimum
    cons.append(c0 >= 0)  # initial fluorescence larger than 0
    cons.append(s >= 0)  # spike train non-negativity
    # noise constraints
    cons_noise = [noise[i] <= thres_sn[i] for i in range(thres_sn.shape[0])]
    try:
        obj = cvx.Minimize(cvx.sum(cvx.norm(s, 1, axis=1)))
        prob = cvx.Problem(obj, cons + cons_noise)
        if use_cons:
            _ = prob.solve(solver="ECOS")
        if not (prob.status == "optimal" or prob.status == "optimal_inaccurate"):
            if use_cons:
                warnings.warn("constrained version of problem infeasible")
            raise ValueError
    except (ValueError, cvx.SolverError):
        lam = (
            sn * sparse_penal / sn.shape[0]
        )  # hacky correction for near-linear relationship between sparsity and number of concurrently updated units
        obj = cvx.Minimize(
            cvx.sum(cvx.sum(noise, axis=1) + lam * cvx.norm(s, 1, axis=1))
        )
        prob = cvx.Problem(obj, cons)
        try:
            _ = prob.solve(solver="ECOS", max_iters=max_iters)
            if prob.status in ["infeasible", "unbounded", None]:
                raise ValueError
        except (cvx.SolverError, ValueError):
            try:
                if scs:
                    _ = prob.solve(solver="SCS", max_iters=200)
                if prob.status in ["infeasible", "unbounded", None]:
                    raise ValueError
            except (cvx.SolverError, ValueError):
                warnings.warn(
                    "problem status is {}, returning null".format(prob.status),
                    RuntimeWarning,
                )
                return np.full((5, c.shape[0], c.shape[1]), np.nan).squeeze()
    if not (prob.status == "optimal"):
        warnings.warn("problem solved sub-optimally", RuntimeWarning)
    return np.stack(
        np.broadcast_arrays(
            c.value, s.value, b.value, c0.value.reshape((-1, 1)), dc_vec
        )
    ).squeeze()


def unit_merge(A, C, add_list=None, thres_corr=0.9):
    """[summary]

    Args:
        A ([type]): [description]
        C ([type]): [description]
        add_list ([type], optional): [description]. Defaults to None.
        thres_corr (float, optional): [description]. Defaults to 0.9.

    Returns:
        [type]: [description]
    """
    print("computing spatial overlap")
    A_bl = (A > 0).astype(np.float32).chunk(dict(unit_id="auto", height=-1, width=-1))
    A_ovlp = xr.apply_ufunc(
        da.array.tensordot,
        A_bl,
        A_bl.rename(unit_id="unit_id_cp"),
        input_core_dims=[
            ["unit_id", "height", "width"],
            ["height", "width", "unit_id_cp"],
        ],
        output_core_dims=[["unit_id", "unit_id_cp"]],
        dask="allowed",
        kwargs=dict(axes=([1, 2], [0, 1])),
        output_dtypes=[A_bl.dtype],
    )
    A_ovlp = A_ovlp.persist()
    print("computing temporal correlation")
    uid_idx = C.coords["unit_id"].values
    corr = xr.apply_ufunc(
        np.corrcoef,
        C.compute(),
        input_core_dims=[["unit_id", "frame"]],
        output_core_dims=[["unit_id", "unit_id_cp"]],
        output_sizes=dict(unit_id_cp=len(uid_idx)),
    )
    corr = corr.assign_coords(unit_id_cp=uid_idx)
    print("labeling units to be merged")
    adj = np.logical_and(A_ovlp > 0, corr > thres_corr)
    unit_labels = xr.apply_ufunc(
        label_connected,
        adj.compute(),
        input_core_dims=[["unit_id", "unit_id_cp"]],
        output_core_dims=[["unit_id"]],
    )
    print("merging units")
    A_merge = (
        A.assign_coords(unit_labels=unit_labels)
        .groupby("unit_labels")
        .sum("unit_id")
        .persist()
        .rename(unit_labels="unit_id")
    )
    C_merge = (
        C.assign_coords(unit_labels=unit_labels)
        .groupby("unit_labels")
        .mean("unit_id")
        .persist()
        .rename(unit_labels="unit_id")
    )
    A_merge = rechunk_like(A_merge, A)
    C_merge = rechunk_like(C_merge, C)
    if add_list:
        for ivar, var in enumerate(add_list):
            var_mrg = (
                var.assign_coords(unit_labels=unit_labels)
                .groupby("unit_labels")
                .mean("unit_id")
                .persist()
                .rename(unit_labels="unit_id")
            )
            add_list[ivar] = rechunk_like(var_mrg, var)
        return A_merge, C_merge, add_list
    else:
        return A_merge, C_merge


def label_connected(adj, only_connected=False):
    """[summary]

    Args:
        adj ([type]): [description]
        only_connected (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    np.fill_diagonal(adj, 0)
    adj = np.triu(adj)
    g = nx.convert_matrix.from_numpy_matrix(adj)
    labels = np.zeros(adj.shape[0], dtype=np.int)
    for icomp, comp in enumerate(nx.connected_components(g)):
        comp = list(comp)
        if only_connected and len(comp) == 1:
            labels[comp] = -1
        else:
            labels[comp] = icomp
    return labels


def smooth_sig(sig, freq, btype="low"):
    """[summary]

    Args:
        sig ([type]): [description]
        freq ([type]): [description]
        btype (str, optional): [description]. Defaults to 'low'.

    Returns:
        [type]: [description]
    """
    but_b, but_a = butter(2, freq, btype=btype, analog=False)
    sig_smth = xr.apply_ufunc(
        lambda x: lfilter(but_b, but_a, x),
        sig.chunk(dict(frame=-1)),
        input_core_dims=[["frame"]],
        output_core_dims=[["frame"]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[sig.dtype],
    )
    return sig_smth



import cv2
import dask.array as darr
import numpy as np
import xarray as xr
from dask import delayed


def estimate_shifts(varr, max_sh, dim="frame", npart=3, local=False):
    """[summary]

    Args:
        varr ([type]): [description]
        max_sh ([type]): [description]
        dim (str, optional): [description]. Defaults to "frame".
        npart (int, optional): [description]. Defaults to 3.
        local (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    varr = varr.chunk({"height": -1, "width": -1})
    vmax, sh = xr.apply_ufunc(
        est_sh_part,
        varr,
        input_core_dims=[[dim, "height", "width"]],
        output_core_dims=[["height", "width"], [dim, "variable"]],
        dask="allowed",
        kwargs={"max_sh": max_sh, "npart": npart, "local": local},
    )
    return sh.assign_coords(variable=["height", "width"])


def est_sh_part(varr, max_sh, npart, local):
    """[summary]

    Args:
        varr ([type]): [description]
        max_sh ([type]): [description]
        npart ([type]): [description]
        local ([type]): [description]

    Returns:
        [type]: [description]
    """
    if varr.shape[0] <= 1:
        return varr.squeeze(), np.array([[0, 0]])
    idx_spt = np.array_split(np.arange(varr.shape[0]), npart)
    fm_ls, sh_ls = [], []
    for idx in idx_spt:
        if len(idx) > 0:
            fm, sh = est_sh_part(varr[idx, :, :], max_sh, npart, local)
            fm_ls.append(fm)
            sh_ls.append(sh)
    mid = int(len(sh_ls) / 2)
    sh_add_ls = [np.array([0, 0])] * len(sh_ls)
    for i, fm in enumerate(fm_ls):
        if i < mid:
            temp = fm_ls[i + 1]
            sh_idx = np.arange(i + 1)
        elif i > mid:
            temp = fm_ls[i - 1]
            sh_idx = np.arange(i, len(sh_ls))
        else:
            continue
        sh_add = darr.from_delayed(
            delayed(match_temp)(fm, temp, max_sh, local), (2,), float
        )
        for j in sh_idx:
            sh_ls[j] = sh_ls[j] + sh_add.reshape((1, -1))
            sh_add_ls[j] = sh_add_ls[j] + sh_add
    for i, (fm, sh) in enumerate(zip(fm_ls, sh_add_ls)):
        fm_ls[i] = darr.nan_to_num(
            darr.from_delayed(delayed(shift_perframe)(fm, sh), fm.shape, fm.dtype)
        )
    sh_ret = darr.concatenate(sh_ls)
    fm_ret = darr.stack(fm_ls)
    return fm_ret.max(axis=0), sh_ret


def match_temp(src, dst, max_sh, local, subpixel=False):
    """[summary]

    Args:
        src ([type]): [description]
        dst ([type]): [description]
        max_sh ([type]): [description]
        local ([type]): [description]
        subpixel (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    src = np.pad(src, max_sh)
    cor = cv2.matchTemplate(
        src.astype(np.float32), dst.astype(np.float32), cv2.TM_CCOEFF_NORMED
    )
    if not len(np.unique(cor)) > 1:
        return np.array([0, 0])
    cent = np.floor(np.array(cor.shape) / 2)
    if local:
        cor_ma = cv2.dilate(cor, np.ones((3, 3)))
        maxs = np.array(np.nonzero(cor_ma == cor))
        dev = ((maxs - cent[:, np.newaxis]) ** 2).sum(axis=0)
        imax = maxs[:, np.argmin(dev)]
    else:
        imax = np.unravel_index(np.argmax(cor), cor.shape)
    if subpixel:
        x0 = np.arange(max(imax[0] - 5, 0), min(imax[0] + 6, cor.shape[0]))
        x1 = np.arange(max(imax[1] - 5, 0), min(imax[1] + 6, cor.shape[1]))
        y0 = cor[x0, imax[1]]
        y1 = cor[imax[0], x1]
        p0 = np.polyfit(x0, y0, 2)
        p1 = np.polyfit(x1, y1, 2)
        imax = np.array([-0.5 * p0[1] / p0[0], -0.5 * p1[1] / p1[0]])
        # m0 = np.roots(np.polyder(p0))
        # m1 = np.roots(np.polyder(p1))
        # m0 = m0[np.argmin(np.abs(m0 - imax[0]))]
        # m1 = m1[np.argmin(np.abs(m1 - imax[1]))]
        # imax = np.array([m0, m1])
    sh = cent - imax
    return sh


def apply_shifts(varr, shifts):
    """[summary]

    Args:
        varr ([type]): [description]
        shifts ([type]): [description]

    Returns:
        [type]: [description]
    """
    sh_dim = shifts.coords["variable"].values.tolist()
    varr_sh = xr.apply_ufunc(
        shift_perframe,
        varr.chunk({d: -1 for d in sh_dim}),
        shifts,
        input_core_dims=[sh_dim, ["variable"]],
        output_core_dims=[sh_dim],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[varr.dtype],
    )
    return varr_sh


def shift_perframe(fm, sh):
    """[summary]

    Args:
        fm ([type]): [description]
        sh ([type]): [description]

    Returns:
        [type]: [description]
    """
    sh = np.around(sh).astype(int)
    fm = np.roll(fm, sh, axis=np.arange(fm.ndim))
    index = [slice(None) for _ in range(fm.ndim)]
    for ish, s in enumerate(sh):
        index = [slice(None) for _ in range(fm.ndim)]
        if s > 0:
            index[ish] = slice(None, s)
            fm[tuple(index)] = np.nan
        elif s == 0:
            continue
        elif s < 0:
            index[ish] = slice(s, None)
            fm[tuple(index)] = np.nan
    return fm



import itertools as itt

import dask as da
import numpy as np
import pandas as pd

from .visualization import centroid


def calculate_centroids(A, window):
    A = A.where(window, 0)
    return centroid(A, verbose=True)


def calculate_centroid_distance(
    cents, by="session", index_dim=["animal"], tile=(50, 50)
):
    res_list = []

    def cent_pair(grp):
        dist_df_ls = []
        len_df = 0
        for (byA, grpA), (byB, grpB) in itt.combinations(list(grp.groupby(by)), 2):
            cur_pairs = subset_pairs(grpA, grpB, tile)
            pairs_ls = list(cur_pairs)
            len_df = len_df + len(pairs_ls)
            subA = grpA.set_index("unit_id").loc[[p[0] for p in pairs_ls]].reset_index()
            subB = grpB.set_index("unit_id").loc[[p[1] for p in pairs_ls]].reset_index()
            dist = da.delayed(pd_dist)(subA, subB).rename("distance")
            dist_df = da.delayed(pd.concat)(
                [subA["unit_id"].rename(byA), subB["unit_id"].rename(byB), dist],
                axis="columns",
            )
            dist_df = dist_df.rename(
                columns={
                    "distance": ("variable", "distance"),
                    byA: (by, byA),
                    byB: (by, byB),
                }
            )
            dist_df_ls.append(dist_df)
        dist_df = da.delayed(pd.concat)(dist_df_ls, ignore_index=True, sort=True)
        return dist_df, len_df

    print("creating parallel schedule")
    if index_dim:
        for idxs, grp in cents.groupby(index_dim):
            dist_df, len_df = cent_pair(grp)
            if type(idxs) is not tuple:
                idxs = (idxs,)
            meta_df = pd.concat(
                [
                    pd.Series([idx] * len_df, name=("meta", dim))
                    for idx, dim in zip(idxs, index_dim)
                ],
                axis="columns",
            )
            res_df = da.delayed(pd.concat)([meta_df, dist_df], axis="columns")
            res_list.append(res_df)
    else:
        res_list = [cent_pair(cents)[0]]
    print("computing distances")
    res_list = da.compute(res_list)[0]
    res_df = pd.concat(res_list, ignore_index=True)
    res_df.columns = pd.MultiIndex.from_tuples(res_df.columns)
    return res_df


def subset_pairs(A, B, tile):
    Ah, Aw, Bh, Bw = A["height"], A["width"], B["height"], B["width"]
    hh = (min(Ah.min(), Bh.min()), max(Ah.max(), Bh.max()))
    ww = (min(Aw.min(), Bw.min()), max(Aw.max(), Bw.max()))
    dh, dw = int(np.ceil(tile[0] / 2)), int(np.ceil(tile[1] / 2))
    tile_h = np.linspace(hh[0], hh[1], int(np.ceil((hh[1] - hh[0]) * 2 / tile[0])))
    tile_w = np.linspace(ww[0], ww[1], int(np.ceil((ww[1] - ww[0]) * 2 / tile[1])))
    pairs = set()
    for h, w in itt.product(tile_h, tile_w):
        curA = A[Ah.between(h - dh, h + dh) & Aw.between(w - dw, w + dw)]
        curB = B[Bh.between(h - dh, h + dh) & Bw.between(w - dw, w + dw)]
        Au, Bu = curA["unit_id"].values, curB["unit_id"].values
        pairs.update(set(map(tuple, cartesian(Au, Bu).tolist())))
    return pairs


def pd_dist(A, B):
    return np.sqrt(
        ((A[["height", "width"]] - B[["height", "width"]]) ** 2).sum("columns")
    )


def cartesian(*args):
    n = len(args)
    return np.array(np.meshgrid(*args)).T.reshape((-1, n))


def group_by_session(df):
    ss = df["session"].notnull()
    grp = ss.apply(lambda r: tuple(r.index[r].tolist()), axis=1)
    df["group", "group"] = grp
    return df


def calculate_mapping(dist):
    map_idxs = set()
    try:
        for anm, grp in dist.groupby(dist["meta", "animal"]):
            map_idxs.update(mapping(grp))
    except KeyError:
        map_idxs = mapping(dist)
    return dist.loc[list(map_idxs)]


def mapping(dist):
    map_list = set()
    for sess, grp in dist.groupby(dist["group", "group"]):
        minidx_list = []
        for ss in sess:
            minidx = set()
            for uid, uid_grp in grp.groupby(grp["session", ss]):
                minidx.add(uid_grp["variable", "distance"].idxmin())
            minidx_list.append(minidx)
        minidxs = set.intersection(*minidx_list)
        map_list.update(minidxs)
    return map_list


def resolve_mapping(mapping):
    map_list = []
    try:
        for anm, grp in mapping.groupby(mapping["meta", "animal"]):
            map_list.append(resolve(grp))
    except KeyError:
        map_list = [resolve(mapping)]
    return pd.concat(map_list, ignore_index=True)


def resolve(mapping):
    mapping = mapping.reset_index(drop=True)
    map_ss = mapping["session"]
    for ss in map_ss.columns:
        del_idx = []
        for ss_uid, ss_grp in mapping.groupby(mapping["session", ss]):
            if ss_grp.shape[0] > 1:
                del_idx.extend(ss_grp.index)
                new_sess = []
                for s in ss_grp["session"]:
                    uval = ss_grp["session", s].dropna().unique()
                    if len(uval) == 0:
                        new_sess.append(np.nan)
                    elif len(uval) == 1:
                        new_sess.append(uval[0])
                    elif len(uval) > 1:
                        break
                else:
                    new_row = ss_grp.iloc[0].copy()
                    new_row["session"] = new_sess
                    new_row["variable", "distance"] = np.nan
                    mapping = mapping.append(new_row, ignore_index=True)
        mapping = mapping.drop(del_idx).reset_index(drop=True)
    return group_by_session(mapping)


def fill_mapping(mappings, cents):
    def fill(cur_grp, cur_cent):
        fill_ls = []
        for cur_ss in list(cur_grp["session"]):
            cur_ss_grp = cur_grp["session"][cur_ss].dropna()
            cur_ss_all = cur_cent[cur_cent["session"] == cur_ss]["unit_id"].dropna()
            cur_fill_set = set(cur_ss_all.unique()) - set(cur_ss_grp.unique())
            cur_fill_df = pd.DataFrame({("session", cur_ss): list(cur_fill_set),})
            cur_fill_df[("group", "group")] = [(cur_ss,)] * len(cur_fill_df)
            fill_ls.append(cur_fill_df)
        return pd.concat(fill_ls, ignore_index=True)

    meta_cols = list(filter(lambda c: c[0] == "meta", mappings.columns))
    if meta_cols:
        meta_cols_smp = [c[1] for c in meta_cols]
        for cur_id, cur_grp in mappings.groupby(meta_cols):
            cur_cent = cents.set_index(meta_cols_smp).loc[cur_id].reset_index()
            cur_grp_fill = fill(cur_grp, cur_cent)
            cur_id = cur_id if type(cur_id) is tuple else tuple([cur_id])
            for icol, col in enumerate(meta_cols):
                cur_grp_fill[col] = cur_id[icol]
            mappings = pd.concat([mappings, cur_grp_fill], ignore_index=True)
    else:
        map_fill = fill(mappings, cents)
        mappings = pd.concat([mappings, map_fill], ignore_index=True)
    return mappings




import cv2
import dask
import dask.array as da
import numpy as np
import pandas as pd
import xarray as xr
from scipy.ndimage.filters import median_filter
from scipy.ndimage.measurements import label
from scipy.signal import butter, hilbert, lfilter
from scipy.spatial.distance import pdist, squareform
from scipy.stats import kstest, zscore
from skimage.morphology import disk
from sklearn.mixture import GaussianMixture

from .cnmf import label_connected, smooth_sig
from .utilities import get_optimal_chk, rechunk_like


def seeds_init(
    varr,
    wnd_size=500,
    method="rolling",
    stp_size=200,
    nchunk=100,
    max_wnd=10,
    diff_thres=2,
):
    """[summary]

    Args:
        varr ([type]): [description]
        wnd_size (int, optional): [description]. Defaults to 500.
        method (str, optional): [description]. Defaults to 'rolling'.
        stp_size (int, optional): [description]. Defaults to 200.
        nchunk (int, optional): [description]. Defaults to 100.
        max_wnd (int, optional): [description]. Defaults to 10.
        diff_thres (int, optional): [description]. Defaults to 2.

    Returns:
        [type]: [description]
    """
    print("constructing chunks")
    idx_fm = varr.coords["frame"]
    nfm = len(idx_fm)
    if method == "rolling":
        nstp = np.ceil(nfm / stp_size) + 1
        centers = np.linspace(0, nfm - 1, int(nstp))
        hwnd = np.ceil(wnd_size / 2)
        max_idx = list(
            map(
                lambda c: slice(
                    int(np.floor(c - hwnd).clip(0)), int(np.ceil(c + hwnd))
                ),
                centers,
            )
        )
    elif method == "random":
        max_idx = [np.random.randint(0, nfm - 1, wnd_size) for _ in range(nchunk)]
    res = []
    print("creating parallel scheme")
    res = [max_proj_frame(varr, cur_idx) for cur_idx in max_idx]
    max_res = xr.concat(res, "sample").chunk(dict(sample=10))
    print("computing max projections")
    max_res = max_res.persist()
    print("calculating local maximum")
    loc_max = xr.apply_ufunc(
        local_max_roll,
        max_res.chunk(dict(height=-1, width=-1)),
        input_core_dims=[["height", "width"]],
        output_core_dims=[["height", "width"]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[np.uint8],
        kwargs=dict(k0=2, k1=max_wnd, diff=diff_thres),
    ).sum("sample")
    loc_max = loc_max.compute()
    loc_max_flt = loc_max.stack(spatial=["height", "width"])
    seeds = (
        loc_max_flt.where(loc_max_flt > 0, drop=True)
        .rename("seeds")
        .to_dataframe()
        .reset_index()
    )
    return seeds[["height", "width", "seeds"]].reset_index()


def max_proj_frame(varr, idx):
    """[summary]

    Args:
        varr ([type]): [description]
        idx ([type]): [description]

    Returns:
        [type]: [description]
    """
    return varr.isel(frame=idx).max("frame")


def local_max_roll(fm, k0, k1, diff):
    """[summary]

    Args:
        fm ([type]): [description]
        k0 ([type]): [description]
        k1 ([type]): [description]
        diff ([type]): [description]

    Returns:
        [type]: [description]
    """
    max_ls = []
    for ksize in range(k0, k1):
        selem = disk(ksize)
        fm_max = local_max(fm, selem, diff)
        max_ls.append(fm_max)
    lmax = (np.stack(max_ls, axis=0).sum(axis=0) > 0).astype(np.uint8)
    nlab, max_lab = cv2.connectedComponents(lmax)
    max_res = np.zeros_like(lmax)
    for lb in range(1, nlab):
        area = max_lab == lb
        if np.sum(area) > 1:
            crds = tuple(int(np.median(c)) for c in np.where(area))
            max_res[crds] = 1
        else:
            max_res[np.where(area)] = 1
    return max_res


def local_max(fm, k, diff=0):
    """[summary]

    Args:
        fm ([type]): [description]
        k ([type]): [description]
        diff (int, optional): [description]. Defaults to 0.

    Returns:
        [type]: [description]
    """
    fm_max = cv2.dilate(fm, k)
    fm_min = cv2.erode(fm, k)
    fm_diff = ((fm_max - fm_min) > diff).astype(np.uint8)
    fm_max = (fm == fm_max).astype(np.uint8)
    return cv2.bitwise_and(fm_max, fm_diff).astype(np.uint8)


def gmm_refine(
    varr, seeds, q=(0.1, 99.9), n_components=2, valid_components=1, mean_mask=True
):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        q (tuple, optional): [description]. Defaults to (0.1, 99.9).
        n_components (int, optional): [description]. Defaults to 2.
        valid_components (int, optional): [description]. Defaults to 1.
        mean_mask (bool, optional): [description]. Defaults to True.

    Returns:
        [type]: [description]
    """
    print("selecting seeds")
    varr_sub = varr.sel(spatial=[tuple(hw) for hw in seeds[["height", "width"]].values])
    print("computing peak-valley values")
    varr_valley = xr.apply_ufunc(
        np.percentile,
        varr_sub.chunk(dict(frame=-1)),
        input_core_dims=[["frame"]],
        kwargs=dict(q=q[0], axis=-1),
        dask="parallelized",
        output_dtypes=[varr_sub.dtype],
    )
    varr_peak = xr.apply_ufunc(
        np.percentile,
        varr_sub.chunk(dict(frame=-1)),
        input_core_dims=[["frame"]],
        kwargs=dict(q=q[1], axis=-1),
        dask="parallelized",
        output_dtypes=[varr_sub.dtype],
    )
    varr_pv = varr_peak - varr_valley
    varr_pv = varr_pv.compute()
    print("fitting GMM models")
    dat = varr_pv.values.reshape(-1, 1)
    gmm = GaussianMixture(n_components=n_components)
    gmm.fit(dat)
    idg = np.argsort(gmm.means_.reshape(-1))[-valid_components:]
    idx_valid = np.isin(gmm.predict(dat), idg)
    if mean_mask:
        idx_mean = dat > np.sort(gmm.means_)[0]
        idx_valid = np.logical_and(idx_mean.squeeze(), idx_valid)
    seeds["mask_gmm"] = idx_valid
    return seeds, varr_pv, gmm


def pnr_refine(varr, seeds, noise_freq=0.25, thres=1.5, q=(0.1, 99.9), med_wnd=None):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        noise_freq (float, optional): [description]. Defaults to 0.25.
        thres (float, optional): [description]. Defaults to 1.5.
        q (tuple, optional): [description]. Defaults to (0.1, 99.9).
        med_wnd ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    print("selecting seeds")
    varr_sub = varr.sel(spatial=[tuple(hw) for hw in seeds[["height", "width"]].values])
    varr_sub = varr_sub.chunk(dict(frame=-1, spatial="auto"))
    if med_wnd:
        varr_base = xr.apply_ufunc(
            median_filter,
            varr_sub,
            input_core_dims=[["frame"]],
            output_core_dims=[["frame"]],
            dask="parallelized",
            kwargs=dict(size=med_wnd),
            vectorize=True,
            output_dtypes=[varr_sub.dtype],
        )
        varr_sub = (varr_sub - varr_base).persist()
    print("computing peak-noise ratio")
    but_b, but_a = butter(2, noise_freq, btype="high", analog=False)
    varr_noise = xr.apply_ufunc(
        lambda x: lfilter(but_b, but_a, x),
        varr_sub.chunk(dict(frame=-1)),
        input_core_dims=[["frame"]],
        output_core_dims=[["frame"]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[varr_sub.dtype],
    )

    def ptp_q(x):
        return np.percentile(x, q[1]) - np.percentile(x, q[0])

    varr_sub_ptp = xr.apply_ufunc(
        ptp_q,
        varr_sub.chunk(dict(frame=-1)),
        input_core_dims=[["frame"]],
        dask="parallelized",
        vectorize=True,
        output_dtypes=[varr_sub.dtype],
    ).compute()
    varr_noise_ptp = xr.apply_ufunc(
        ptp_q,
        varr_noise.chunk(dict(frame=-1)).real,
        input_core_dims=[["frame"]],
        dask="parallelized",
        vectorize=True,
        output_dtypes=[varr_sub.dtype],
    ).compute()
    pnr = varr_sub_ptp / varr_noise_ptp
    if thres == "auto":
        gmm = GaussianMixture(n_components=2)
        gmm.fit(np.nan_to_num(pnr.values.reshape(-1, 1)))
        idg = np.argsort(gmm.means_.reshape(-1))[-1]
        idx_valid = np.isin(gmm.predict(pnr.values.reshape(-1, 1)), idg)
        seeds["mask_pnr"] = idx_valid
    else:
        mask = pnr > thres
        mask = mask.compute()
        mask_df = mask.to_pandas().rename("mask_pnr").reset_index()
        seeds = pd.merge(seeds, mask_df, on=["height", "width"], how="left")
        gmm = None
    return seeds, pnr, gmm


def intensity_refine(varr, seeds, thres_mul=2):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        thres_mul (int, optional): [description]. Defaults to 2.

    Returns:
        [type]: [description]
    """
    try:
        fm_max = varr.max("frame")
    except ValueError:
        print("using input as max projection")
        fm_max = varr
    bins = np.around(fm_max.sizes["height"] * fm_max.sizes["width"] / 10).astype(int)
    hist, edges = np.histogram(fm_max, bins=bins)
    try:
        thres = edges[int(np.around(np.argmax(hist) * thres_mul))]
    except IndexError:
        print("threshold out of bound, returning input")
        return seeds
    mask = (fm_max > thres).stack(spatial=["height", "width"])
    mask_df = mask.to_pandas().rename("mask_int").reset_index()
    seeds = pd.merge(seeds, mask_df, on=["height", "width"], how="left")
    return seeds


def ks_refine(varr, seeds, sig=0.05):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        sig (float, optional): [description]. Defaults to 0.05.

    Returns:
        [type]: [description]
    """
    print("selecting seeds")
    varr_sub = varr.sel(spatial=[tuple(hw) for hw in seeds[["height", "width"]].values])
    print("performing KS test")
    ks = xr.apply_ufunc(
        lambda x: kstest(zscore(x), "norm")[1],
        varr_sub.chunk(dict(frame=-1, spatial="auto")),
        input_core_dims=[["frame"]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    )
    mask = ks < sig
    mask_df = mask.to_pandas().rename("mask_ks").reset_index()
    seeds = pd.merge(seeds, mask_df, on=["height", "width"], how="left")
    return seeds


def seeds_merge(varr, seeds, thres_dist=5, thres_corr=0.6, noise_freq="envelope"):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        thres_dist (int, optional): [description]. Defaults to 5.
        thres_corr (float, optional): [description]. Defaults to 0.6.
        noise_freq (str, optional): [description]. Defaults to 'envelope'.

    Returns:
        [type]: [description]
    """
    crds = [tuple([h, w]) for h, w in seeds[["height", "width"]].values]
    nsmp = len(crds)
    varr_sub = varr.sel(spatial=crds)
    varr_max = varr_sub.max("frame").compute()
    print("computing distance")
    dist = xr.DataArray(
        squareform(pdist(seeds[["height", "width"]].values)),
        dims=["sampleA", "sampleB"],
        coords=dict(sampleA=np.arange(nsmp), sampleB=np.arange(nsmp)),
    )
    if noise_freq:
        if noise_freq == "envelope":
            print("computing hilbert transform")
            varr_sub = xr.apply_ufunc(
                lambda x: abs(hilbert(x)),
                varr_sub.chunk(dict(frame=-1, spatial="auto")),
                input_core_dims=[["frame"]],
                output_core_dims=[["frame"]],
                vectorize=True,
                output_dtypes=[varr_sub.dtype],
                dask="parallelized",
            )
        else:
            varr_sub = smooth_sig(varr_sub, noise_freq)
    corr = xr.apply_ufunc(
        da.corrcoef,
        varr_sub.chunk(dict(spatial="auto", frame=-1)),
        input_core_dims=[["spatial", "frame"]],
        output_core_dims=[["sampleA", "sampleB"]],
        dask="allowed",
        output_sizes=dict(sampleA=nsmp, sampleB=nsmp),
        output_dtypes=[float],
    ).assign_coords(sampleA=np.arange(nsmp), sampleB=np.arange(nsmp))
    print("computing correlations")
    corr = corr.compute()
    adj = np.logical_and(dist < thres_dist, corr > thres_corr)
    adj = adj.compute()
    np.fill_diagonal(adj.values, 0)
    iso = adj.sum("sampleB")
    iso = iso.where(iso == 0).dropna("sampleA")
    labels = label_connected(adj.values)
    uids = adj.coords["sampleA"].values
    seeds_final = set(iso.coords["sampleA"].data.tolist())
    for cur_cmp in np.unique(labels):
        cur_smp = uids[np.where(labels == cur_cmp)[0]]
        cur_max = varr_max.isel(spatial=cur_smp)
        max_seed = cur_smp[np.argmax(cur_max.data)]
        seeds_final.add(max_seed)
    seeds["mask_mrg"] = False
    seeds.loc[list(seeds_final), "mask_mrg"] = True
    return seeds


def initialize(varr, seeds, thres_corr=0.8, wnd=10, noise_freq=None):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        thres_corr (float, optional): [description]. Defaults to 0.8.
        wnd (int, optional): [description]. Defaults to 10.
        noise_freq ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    print("creating parallel schedule")
    harr, warr = seeds["height"].values, seeds["width"].values
    varr_rechk = varr.chunk(dict(frame=-1))
    res_ls = [
        init_perseed(varr_rechk, h, w, wnd, thres_corr, noise_freq)
        for h, w in zip(harr, warr)
    ]
    print("computing ROIs")
    res_ls = dask.compute(res_ls)[0]
    print("concatenating results")
    A = xr.concat([r[0] for r in res_ls], "unit_id").assign_coords(
        unit_id=np.arange(len(res_ls))
    )
    C = xr.concat([r[1] for r in res_ls], "unit_id").assign_coords(
        unit_id=np.arange(len(res_ls))
    )
    print("initializing backgrounds")
    A = A.reindex_like(varr.isel(frame=0)).fillna(0)
    chk = {d: c for d, c in zip(varr.dims, varr.chunks)}
    uchkA = get_optimal_chk(varr, A)["unit_id"]
    uchkC = get_optimal_chk(varr, C)["unit_id"]
    uchk = min(uchkA, uchkC)
    A = A.chunk(dict(height=chk["height"], width=chk["width"], unit_id=uchk))
    C = C.chunk(dict(frame=chk["frame"], unit_id=uchk))
    A_mask = A.sum("unit_id") == 0
    Yb = varr.where(A_mask, 0)
    b = Yb.mean("frame").persist()
    f = Yb.mean(["height", "width"]).persist()
    b = rechunk_like(b, varr)
    return A, C, b, f


def init_perseed(varr, h, w, wnd, thres_corr, noise_freq):
    """[summary]

    Args:
        varr ([type]): [description]
        h ([type]): [description]
        w ([type]): [description]
        wnd ([type]): [description]
        thres_corr ([type]): [description]
        noise_freq ([type]): [description]

    Returns:
        [type]: [description]
    """
    ih = np.where(varr.coords["height"] == h)[0][0]
    iw = np.where(varr.coords["width"] == w)[0][0]
    h_sur, w_sur = (
        slice(max(ih - wnd, 0), ih + wnd),
        slice(max(iw - wnd, 0), iw + wnd),
    )
    sur = varr.isel(height=h_sur, width=w_sur)
    sur_flt = sur.stack(spatial=["height", "width"])
    ih = np.where(sur.coords["height"] == h)[0][0]
    iw = np.where(sur.coords["width"] == w)[0][0]
    sp_idxs = sur_flt.coords["spatial"].values
    if noise_freq:
        sur_smth = smooth_sig(sur_flt, noise_freq)
    else:
        sur_smth = sur_flt
    corr = xr.apply_ufunc(
        da.corrcoef,
        sur_smth,
        input_core_dims=[["spatial", "frame"]],
        output_core_dims=[["spatial", "spatial_cp"]],
        dask="allowed",
        output_sizes=dict(spatial_cp=len(sp_idxs)),
    )
    sd_id = np.ravel_multi_index((ih, iw), (sur.sizes["height"], sur.sizes["width"]))
    corr = corr.isel(spatial_cp=sd_id).squeeze().unstack("spatial")
    mask = corr > thres_corr
    mask_lb = xr.apply_ufunc(da_label, mask, dask="allowed")
    sd_lb = mask_lb.isel(height=ih, width=iw)
    mask = mask_lb == sd_lb
    sur = sur.where(mask, 0)
    corr = corr.where(mask, 0)
    corr_norm = corr / corr.sum()
    C = xr.apply_ufunc(
        da.tensordot,
        sur,
        corr_norm,
        input_core_dims=[["frame", "height", "width"], ["height", "width"]],
        output_core_dims=[["frame"]],
        kwargs=dict(axes=[(1, 2), (0, 1)]),
        dask="allowed",
    )
    return corr, C


@da.as_gufunc(signature="(h, w)->(h, w)", output_dtypes=int, allow_rechunk=True)
def da_label(im):
    """[summary]

    Args:
        im ([type]): [description]

    Returns:
        [type]: [description]
    """
    return label(im)[0]



import functools as fct
import itertools as itt
import os
from collections import OrderedDict
from uuid import uuid4

import av
import cv2
import dask
import dask.array as da
import ffmpeg
import holoviews as hv
import numpy as np
import pandas as pd
import panel as pn
import param
import xarray as xr
from bokeh.palettes import Category10_10, Viridis256
from dask.diagnostics import ProgressBar
from datashader import count_cat
from holoviews.operation.datashader import datashade, dynspread, regrid
from holoviews.streams import (
    BoxEdit,
    DoubleTap,
    Pipe,
    RangeXY,
    Selection1D,
    Stream,
    Tap,
)
from holoviews.util import Dynamic
from panel import widgets as pnwgt
from scipy import linalg
from scipy.ndimage.measurements import center_of_mass
from scipy.spatial import cKDTree


class VArrayViewer:
    """[summary]
    """

    def __init__(
        self,
        varr,
        framerate=30,
        rerange=None,
        summary=["mean"],
        meta_dims=None,
        datashading=True,
        layout=False,
    ):
        """[summary]

        Args:
            varr ([type]): [description]
            framerate (int, optional): [description]. Defaults to 30.
            rerange ([type], optional): [description]. Defaults to None.
            summary (list, optional): [description]. Defaults to ["mean"].
            meta_dims ([type], optional): [description]. Defaults to None.
            datashading (bool, optional): [description]. Defaults to True.
            layout (bool, optional): [description]. Defaults to False.

        Raises:
            NotImplementedError: [description]
        """
        if isinstance(varr, list):
            for iv, v in enumerate(varr):
                varr[iv] = v.assign_coords(data_var=v.name)
            self.ds = xr.concat(varr, dim="data_var")
            meta_dims = ["data_var"]
        elif isinstance(varr, xr.DataArray):
            self.ds = varr.to_dataset()
        elif isinstance(varr, xr.Dataset):
            self.ds = varr
        else:
            raise NotImplementedError(
                "video array of type {} not supported".format(type(varr))
            )
        try:
            self.meta_dicts = OrderedDict(
                [(d, list(self.ds.coords[d].values)) for d in meta_dims]
            )
            self.cur_metas = OrderedDict(
                [(d, v[0]) for d, v in self.meta_dicts.items()]
            )
        except TypeError:
            self.meta_dicts = dict()
            self.cur_metas = dict()
        self._datashade = datashading
        self._layout = layout
        self.framerate = framerate
        self._f = self.ds.coords["frame"].values
        self._h = self.ds.sizes["height"]
        self._w = self.ds.sizes["width"]
        self.mask = dict()
        self.rerange = rerange
        CStream = Stream.define(
            "CStream",
            f=param.Integer(
                default=int(self._f.min()), bounds=(self._f.min(), self._f.max())
            ),
        )
        self.strm_f = CStream()
        self.str_box = BoxEdit()
        self.widgets = self._widgets()
        if type(summary) is list:
            summ_all = {
                "mean": self.ds.mean(["height", "width"]),
                "max": self.ds.max(["height", "width"]),
                "min": self.ds.min(["height", "width"]),
                "diff": self.ds.diff("frame").mean(["height", "width"]),
            }
            try:
                summ = {k: summ_all[k] for k in summary}
            except KeyError:
                print("{} Not understood for specifying summary".format(summary))
            if summ:
                print("computing summary")
                sum_list = []
                for k, v in summ.items():
                    sum_list.append(v.compute().assign_coords(sum_var=k))
                summary = xr.concat(sum_list, dim="sum_var")
        self.summary = summary
        if layout:
            self.ds_sub = self.ds
            self.sum_sub = self.summary
        else:
            self.ds_sub = self.ds.sel(**self.cur_metas)
            try:
                self.sum_sub = self.summary.sel(**self.cur_metas)
            except AttributeError:
                self.sum_sub = self.summary
        self.pnplot = pn.panel(self.get_hvobj())

    def get_hvobj(self):
        """[summary]
        """

        def get_im_ovly(meta):
            """[summary]

            Args:
                meta ([type]): [description]
            """

            def img(f, ds):
                """[summary]

                Args:
                    f ([type]): [description]
                    ds ([type]): [description]

                Returns:
                    [type]: [description]
                """
                return hv.Image(ds.sel(frame=f).compute(), kdims=["width", "height"])

            try:
                curds = self.ds_sub.sel(**meta).rename("_".join(meta.values()))
            except ValueError:
                curds = self.ds_sub
            fim = fct.partial(img, ds=curds)
            im = regrid(hv.DynamicMap(fim, streams=[self.strm_f])).opts(
                frame_width=500, aspect=self._w / self._h, cmap="Viridis"
            )
            if self.rerange:
                im = im.redim.range(**{vname: self.rerange})
            self.xyrange = RangeXY(source=im).rename(x_range="w", y_range="h")
            if not self._layout:
                hv_box = hv.Polygons([]).opts(
                    style={"fill_alpha": 0.3, "line_color": "white"}
                )
                self.str_box = BoxEdit(source=hv_box)
                im_ovly = im * hv_box
            else:
                im_ovly = im

            def hist(f, w, h, ds):
                if w and h:
                    cur_im = hv.Image(
                        ds.sel(frame=f).compute(), kdims=["width", "height"]
                    ).select(height=h, width=w)
                else:
                    cur_im = hv.Image(
                        ds.sel(frame=f).compute(), kdims=["width", "height"]
                    )
                return hv.operation.histogram(cur_im, num_bins=50).opts(
                    xlabel="fluorescence", ylabel="freq"
                )

            fhist = fct.partial(hist, ds=curds)
            his = hv.DynamicMap(fhist, streams=[self.strm_f, self.xyrange]).opts(
                frame_height=int(500 * self._h / self._w), width=150, cmap="Viridis"
            )
            im_ovly = (im_ovly << his).map(lambda p: p.opts(style=dict(cmap="Viridis")))
            return im_ovly

        if self._layout and self.meta_dicts:
            im_dict = OrderedDict()
            for meta in itt.product(*list(self.meta_dicts.values())):
                mdict = {k: v for k, v in zip(list(self.meta_dicts.keys()), meta)}
                im_dict[meta] = get_im_ovly(mdict)
            ims = hv.NdLayout(im_dict, kdims=list(self.meta_dicts.keys()))
        else:
            ims = get_im_ovly(self.cur_metas)
        if self.summary is not None:
            hvsum = (
                hv.Dataset(self.sum_sub)
                .to(hv.Curve, kdims=["frame"])
                .overlay("sum_var")
            )
            if self._datashade:
                hvsum = datashade_ndcurve(hvsum, kdim="sum_var")
            try:
                hvsum = hvsum.layout(list(self.meta_dicts.keys()))
            except:
                pass
            vl = hv.DynamicMap(lambda f: hv.VLine(f), streams=[self.strm_f]).opts(
                style=dict(color="red")
            )
            summ = (hvsum * vl).map(
                lambda p: p.opts(frame_width=500, aspect=3), [hv.RGB, hv.Curve]
            )
        else:
            summ = hv.Div("")
        hvobj = (ims + summ).cols(1)
        return hvobj

    def show(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        return pn.layout.Column(self.widgets, self.pnplot)

    def _widgets(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        w_play = pnwgt.Player(
            length=len(self._f), interval=10, value=0, width=650, height=90
        )

        def play(f):
            if not f.old == f.new:
                self.strm_f.event(f=int(self._f[f.new]))

        w_play.param.watch(play, "value")
        w_box = pnwgt.Button(
            name="Update Mask", button_type="primary", width=100, height=30
        )
        w_box.param.watch(self._update_box, "clicks")
        if not self._layout:
            wgt_meta = {
                d: pnwgt.Select(name=d, options=v, height=45, width=120)
                for d, v in self.meta_dicts.items()
            }

            def make_update_func(meta_name):
                def _update(x):
                    self.cur_metas[meta_name] = x.new
                    self._update_subs()

                return _update

            for d, wgt in wgt_meta.items():
                cur_update = make_update_func(d)
                wgt.param.watch(cur_update, "value")
            wgts = pn.layout.WidgetBox(w_box, w_play, *list(wgt_meta.values()))
        else:
            wgts = pn.layout.WidgetBox(w_box, w_play)
        return wgts

    def _update_subs(self):
        """[summary]
        """
        self.ds_sub = self.ds.sel(**self.cur_metas)
        if self.sum_sub is not None:
            self.sum_sub = self.summary.sel(**self.cur_metas)
        self.pnplot.objects[0].object = self.get_hvobj()

    def _update_box(self, click):
        """[summary]

        Args:
            click ([type]): [description]
        """
        box = self.str_box.data
        self.mask.update(
            {
                tuple(self.cur_metas.values()): {
                    "height": slice(box["y0"][0], box["y1"][0]),
                    "width": slice(box["x0"][0], box["x1"][0]),
                }
            }
        )


class CNMFViewer:
    """[summary]
    """

    def __init__(self, minian=None, A=None, C=None, S=None, org=None, sortNN=True):
        """[summary]

        Args:
            minian ([type], optional): [description]. Defaults to None.
            A ([type], optional): [description]. Defaults to None.
            C ([type], optional): [description]. Defaults to None.
            S ([type], optional): [description]. Defaults to None.
            org ([type], optional): [description]. Defaults to None.
            sortNN (bool, optional): [description]. Defaults to True.
        """
        self._A = A if A is not None else minian["A"]
        self._C = C if C is not None else minian["C"]
        self._S = S if S is not None else minian["S"]
        self._org = org if org is not None else minian["org"]
        try:
            self.unit_labels = minian["unit_labels"].compute()
        except:
            self.unit_labels = xr.DataArray(
                self._A["unit_id"].values.copy(),
                dims=self._A["unit_id"].dims,
                coords=self._A["unit_id"].coords,
            ).rename("unit_labels")
        self._C_norm = xr.apply_ufunc(
            normalize,
            self._C.chunk(dict(frame=-1, unit_id="auto")),
            input_core_dims=[["frame"]],
            output_core_dims=[["frame"]],
            vectorize=True,
            dask="parallelized",
            output_dtypes=[self._C.dtype],
        )
        self._S_norm = xr.apply_ufunc(
            normalize,
            self._S.chunk(dict(frame=-1, unit_id="auto")),
            input_core_dims=[["frame"]],
            output_core_dims=[["frame"]],
            vectorize=True,
            dask="parallelized",
            output_dtypes=[self._C.dtype],
        )
        self.cents = centroid(self._A, verbose=True)
        print("computing sum projection")
        with ProgressBar():
            self.Asum = self._A.sum("unit_id").compute()
        self._NNsort = sortNN
        self._normalize = False
        self._useAC = True
        self._showC = True
        self._showS = True
        meta_dims = list(set(self._org.dims) - {"frame", "height", "width"})
        self.meta_dicts = {d: list(self._org.coords[d].values) for d in meta_dims}
        self.metas = {d: v[0] for d, v in self.meta_dicts.items()}
        if self._NNsort:
            try:
                self.cents["NNord"] = self.cents.groupby(
                    meta_dims, group_keys=False
                ).apply(NNsort)
            except ValueError:
                self.cents["NNord"] = NNsort(self.cents)
            NNcoords = self.cents.set_index(meta_dims + ["unit_id"])[
                "NNord"
            ].to_xarray()
            self._A = self._A.assign_coords(NNord=NNcoords)
            self._C = self._C.assign_coords(NNord=NNcoords)
            self._S = self._S.assign_coords(NNord=NNcoords)
            self._C_norm = self._C_norm.assign_coords(NNord=NNcoords)
            self._S_norm = self._S_norm.assign_coords(NNord=NNcoords)
        self.update_subs()
        self.strm_f = DoubleTap(rename=dict(x="f"))
        self.strm_f.add_subscriber(self.callback_f)
        self.strm_uid = Selection1D()
        self.strm_uid.add_subscriber(self.callback_uid)
        Stream_usub = Stream.define("Stream_usub", usub=param.List())
        self.strm_usub = Stream_usub()
        self.strm_usub.add_subscriber(self.callback_usub)
        self.usub_sel = self.strm_usub.usub
        self._AC = self._org.sel(**self.metas)
        self._mov = self._org.sel(**self.metas)
        self.pipAC = Pipe([])
        self.pipmov = Pipe([])
        self.pipusub = Pipe([])
        self.wgt_meta = self._meta_wgt()
        self.wgt_spatial_all = self._spatial_all_wgt()
        self.spatial_all = self._spatial_all()
        self.temp_comp_sub = self._temp_comp_sub(self._u[:5])
        self.wgt_man = self._man_wgt()
        self.wgt_temp_comp = self._temp_comp_wgt()

    def update_subs(self):
        """[summary]
        """
        self.A_sub = self._A.sel(**self.metas)
        self.C_sub = self._C.sel(**self.metas)
        self.S_sub = self._S.sel(**self.metas)
        self.org_sub = self._org.sel(**self.metas)
        self.C_norm_sub = self._C_norm.sel(**self.metas)
        self.S_norm_sub = self._S_norm.sel(**self.metas)
        if self._NNsort:
            self.A_sub = self.A_sub.sortby("NNord")
            self.C_sub = self.C_sub.sortby("NNord")
            self.S_sub = self.S_sub.sortby("NNord")
            self.C_norm_sub = self.C_norm_sub.sortby("NNord")
            self.S_norm_sub = self.S_norm_sub.sortby("NNord")
        self._h = (
            self.A_sub.isel(unit_id=0)
            .dropna("height", how="all")
            .coords["height"]
            .values
        )
        self._w = (
            self.A_sub.isel(unit_id=0).dropna("width", how="all").coords["width"].values
        )
        self._f = self.C_sub.isel(unit_id=0).dropna("frame").coords["frame"].values
        self._u = self.C_sub.isel(frame=0).dropna("unit_id").coords["unit_id"].values
        if self.meta_dicts:
            sub = pd.concat(
                [self.cents[d] == v for d, v in self.metas.items()], axis="columns"
            ).all(axis="columns")
            self.cents_sub = self.cents[sub]
        else:
            self.cents_sub = self.cents

    def compute_subs(self, clicks=None):
        """[summary]

        Args:
            clicks ([type], optional): [description]. Defaults to None.
        """
        self.A_sub = self.A_sub.compute()
        self.C_sub = self.C_sub.compute()
        self.S_sub = self.S_sub.compute()
        self.org_sub = self.org_sub.compute()
        self.C_norm_sub = self.C_norm_sub.compute()
        self.S_norm_sub = self.S_norm_sub.compute()

    def update_all(self, clicks=None):
        """[summary]

        Args:
            clicks ([type], optional): [description]. Defaults to None.
        """
        self.update_subs()
        self.strm_uid.event(index=[])
        self.strm_f.event(x=0)
        self.update_spatial_all()

    def callback_uid(self, index=None):
        """[summary]

        Args:
            index ([type], optional): [description]. Defaults to None.
        """
        self.update_temp()
        self.update_AC()
        self.update_usub_lab()

    def callback_f(self, f, y):
        """[summary]

        Args:
            f ([type]): [description]
            y ([type]): [description]
        """
        if len(self._AC) > 0 and len(self._mov) > 0:
            fidx = np.abs(self._f - f).argmin()
            f = self._f[fidx]
            if self._useAC:
                AC = self._AC.sel(frame=f)
            else:
                AC = self._AC
            mov = self._mov.sel(frame=f)
            self.pipAC.send(AC)
            self.pipmov.send(mov)
            try:
                self.wgt_temp_comp[1].value = int(fidx)
            except AttributeError:
                pass
        else:
            self.pipAC.send([])
            self.pipmov.send([])

    def callback_usub(self, usub=None):
        """[summary]

        Args:
            usub ([type], optional): [description]. Defaults to None.
        """
        self.update_temp_comp_sub(usub)
        self.update_AC(usub)
        self.update_usub_lab(usub)

    def _meta_wgt(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        wgt_meta = {
            d: pnwgt.Select(name=d, options=v, height=45, width=120)
            for d, v in self.meta_dicts.items()
        }

        def make_update_func(meta_name):
            def _update(x):
                self.metas[meta_name] = x.new
                self.update_subs()

            return _update

        for d, wgt in wgt_meta.items():
            cur_update = make_update_func(d)
            wgt.param.watch(cur_update, "value")
        wgt_update = pnwgt.Button(
            name="Refresh", button_type="primary", height=30, width=120
        )
        wgt_update.param.watch(self.update_all, "clicks")
        wgt_load = pnwgt.Button(
            name="Load Data", button_type="danger", height=30, width=120
        )
        wgt_load.param.watch(self.compute_subs, "clicks")
        return pn.layout.WidgetBox(
            *(list(wgt_meta.values()) + [wgt_update, wgt_load]), width=150
        )

    def show(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        return pn.layout.Column(
            self.spatial_all,
            pn.layout.Row(
                pn.layout.Column(
                    pn.layout.Row(self.wgt_meta, self.wgt_spatial_all),
                    self.wgt_temp_comp,
                ),
                self.temp_comp_sub,
                self.wgt_man,
            ),
        )

    def _temp_comp_sub(self, usub=None):
        """[summary]

        Args:
            usub ([type], optional): [description]. Defaults to None.

        Returns:
            [type]: [description]
        """
        if usub is None:
            usub = self.strm_usub.usub
        if self._normalize:
            C, S = self.C_norm_sub, self.S_norm_sub
        else:
            C, S = self.C_sub, self.S_sub
        cur_temp = dict()
        if self._showC:
            cur_temp["C"] = hv.Dataset(
                C.sel(unit_id=usub)
                .compute()
                .rename("Intensity (A. U.)")
                .dropna("frame", how="all")
            ).to(hv.Curve, "frame")
        if self._showS:
            cur_temp["S"] = hv.Dataset(
                S.sel(unit_id=usub)
                .compute()
                .rename("Intensity (A. U.)")
                .dropna("frame", how="all")
            ).to(hv.Curve, "frame")
        cur_vl = hv.DynamicMap(
            lambda f, y: hv.VLine(f) if f else hv.VLine(0), streams=[self.strm_f]
        ).opts(style=dict(color="red"))
        cur_cv = hv.Curve([], kdims=["frame"], vdims=["Internsity (A.U.)"])
        self.strm_f.source = cur_cv
        h_cv = len(self._w) // 8
        w_cv = len(self._w) * 2
        temp_comp = (
            cur_cv
            * datashade_ndcurve(
                hv.HoloMap(cur_temp, "trace")
                .collate()
                .overlay("trace")
                .grid("unit_id")
                .add_dimension("time", 0, 0),
                "trace",
            )
            .opts(plot=dict(shared_xaxis=True))
            .map(
                lambda p: p.opts(plot=dict(frame_height=h_cv, frame_width=w_cv)), hv.RGB
            )
            * cur_vl
        )
        temp_comp[temp_comp.keys()[0]] = temp_comp[temp_comp.keys()[0]].opts(
            plot=dict(height=h_cv + 75)
        )
        return pn.panel(temp_comp)

    def update_temp_comp_sub(self, usub=None):
        """[summary]

        Args:
            usub ([type], optional): [description]. Defaults to None.
        """
        self.temp_comp_sub.object = self._temp_comp_sub(usub).object
        self.wgt_man.objects = self._man_wgt().objects

    def update_norm(self, norm):
        """[summary]

        Args:
            norm ([type]): [description]
        """
        self._normalize = norm.new
        self.update_temp_comp_sub()

    def _temp_comp_wgt(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        if self.strm_uid.index:
            cur_idxs = self.strm_uid.index
        else:
            cur_idxs = self._u
        ntabs = np.ceil(len(cur_idxs) / 5)
        sub_idxs = np.array_split(cur_idxs, ntabs)
        idxs_dict = OrderedDict(
            [("group{}".format(i), g.tolist()) for i, g in enumerate(sub_idxs)]
        )
        def_idxs = list(idxs_dict.values())[0]
        wgt_grp = pnwgt.Select(
            name="", options=idxs_dict, width=120, height=30, value=def_idxs
        )

        def update_usub(usub):
            self.usub_sel = []
            self.strm_usub.event(usub=usub.new)

        wgt_grp.param.watch(update_usub, "value")
        wgt_grp.value = def_idxs
        self.strm_usub.event(usub=def_idxs)
        wgt_grp_prv = pnwgt.Button(
            name="Previous Group", width=120, height=30, button_type="primary"
        )

        def prv(clicks):
            cur_val = wgt_grp.value
            ig = list(idxs_dict.values()).index(cur_val)
            try:
                prv_val = idxs_dict[list(idxs_dict.keys())[ig - 1]]
                wgt_grp.value = prv_val
            except:
                pass

        wgt_grp_prv.param.watch(prv, "clicks")
        wgt_grp_nxt = pnwgt.Button(
            name="Next Group", width=120, height=30, button_type="primary"
        )

        def nxt(clicks):
            cur_val = wgt_grp.value
            ig = list(idxs_dict.values()).index(cur_val)
            try:
                nxt_val = idxs_dict[list(idxs_dict.keys())[ig + 1]]
                wgt_grp.value = nxt_val
            except:
                pass

        wgt_grp_nxt.param.watch(nxt, "clicks")
        wgt_norm = pnwgt.Checkbox(
            name="Normalize", value=self._normalize, width=120, height=10
        )
        wgt_norm.param.watch(self.update_norm, "value")
        wgt_showC = pnwgt.Checkbox(
            name="ShowC", value=self._showC, width=120, height=10
        )

        def callback_showC(val):
            self._showC = val.new
            self.update_temp_comp_sub()

        wgt_showC.param.watch(callback_showC, "value")
        wgt_showS = pnwgt.Checkbox(
            name="ShowS", value=self._showS, width=120, height=10
        )

        def callback_showS(val):
            self._showS = val.new
            self.update_temp_comp_sub()

        wgt_showS.param.watch(callback_showS, "value")
        wgt_play = pnwgt.Player(length=len(self._f), interval=10, value=0, width=280)

        def play(f):
            if not f.old == f.new:
                self.strm_f.event(x=self._f[f.new])

        wgt_play.param.watch(play, "value")
        wgt_groups = pn.layout.Row(
            pn.layout.WidgetBox(wgt_norm, wgt_showC, wgt_showS, wgt_grp, width=150),
            pn.layout.WidgetBox(wgt_grp_prv, wgt_grp_nxt, width=150),
        )
        return pn.layout.Column(wgt_groups, wgt_play)

    def _man_wgt(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        usub = self.strm_usub.usub
        usub.sort()
        usub.reverse()
        ulabs = self.unit_labels.sel(unit_id=usub).values
        wgt_sel = {
            uid: pnwgt.Select(
                name="Unit Label",
                options=usub + [-1] + ulabs.tolist(),
                value=ulb,
                height=50,
                width=100,
            )
            for uid, ulb in zip(usub, ulabs)
        }

        def callback_ulab(value, uid):
            self.unit_labels.loc[uid] = value.new

        for uid, sel in wgt_sel.items():
            cb = fct.partial(callback_ulab, uid=uid)
            sel.param.watch(cb, "value")
        wgt_check = {
            uid: pnwgt.Checkbox(
                name="Unit ID: {}".format(uid), value=False, height=50, width=100
            )
            for uid in usub
        }

        def callback_chk(val, uid):
            if not val.old == val.new:
                if val.new:
                    self.usub_sel.append(uid)
                else:
                    self.usub_sel.remove(uid)

        for uid, chk in wgt_check.items():
            cb = fct.partial(callback_chk, uid=uid)
            chk.param.watch(cb, "value")
        wgt_discard = pnwgt.Button(
            name="Discard Selected", button_type="primary", width=200
        )

        def callback_discard(clicks):
            for uid in self.usub_sel:
                wgt_sel[uid].value = -1

        wgt_discard.param.watch(callback_discard, "clicks")
        wgt_merge = pnwgt.Button(
            name="Merge Selected", button_type="primary", width=200
        )

        def callback_merge(clicks):
            for uid in self.usub_sel:
                wgt_sel[uid].value = self.usub_sel[0]

        wgt_merge.param.watch(callback_merge, "clicks")
        return pn.layout.Column(
            pn.layout.WidgetBox(wgt_discard, wgt_merge, width=200),
            pn.layout.Row(
                pn.layout.WidgetBox(*wgt_check.values(), width=100),
                pn.layout.WidgetBox(*wgt_sel.values(), width=100),
            ),
        )

    def update_temp_comp_wgt(self):
        """[summary]
        """
        self.wgt_temp_comp.objects = self._temp_comp_wgt().objects

    def update_temp(self):
        """[summary]
        """
        self.update_temp_comp_wgt()

    def update_AC(self, usub=None):
        """[summary]

        Args:
            usub ([type], optional): [description]. Defaults to None.
        """
        if usub is None:
            usub = self.strm_usub.usub
        if usub:
            if self._useAC:
                umask = (self.A_sub.sel(unit_id=usub) > 0).any("unit_id")
                A_sub = self.A_sub.sel(unit_id=usub).where(umask, drop=True).fillna(0)
                C_sub = self.C_sub.sel(unit_id=usub)
                AC = xr.apply_ufunc(
                    da.dot,
                    A_sub,
                    C_sub,
                    input_core_dims=[
                        ["height", "width", "unit_id"],
                        ["unit_id", "frame"],
                    ],
                    output_core_dims=[["height", "width", "frame"]],
                    dask="allowed",
                )
                self._AC = AC.compute()
                wndh, wndw = AC.coords["height"].values, AC.coords["width"].values
                window = self.A_sub.sel(
                    height=slice(wndh.min(), wndh.max()),
                    width=slice(wndw.min(), wndw.max()),
                )
                self._AC = self._AC.reindex_like(window).fillna(0)
                self._mov = (self.org_sub.reindex_like(window)).compute()
            else:
                self._AC = self.A_sub.sel(unit_id=usub).sum("unit_id")
                self._mov = self.org_sub
            self.strm_f.event(x=0)
        else:
            self._AC = xr.DataArray([])
            self._mov = xr.DataArray([])
            self.strm_f.event(x=0)

    def update_usub_lab(self, usub=None):
        """[summary]

        Args:
            usub ([type], optional): [description]. Defaults to None.
        """
        if usub is None:
            usub = self.strm_usub.usub
        if usub:
            self.pipusub.send(self.cents_sub[self.cents_sub["unit_id"].isin(usub)])
        else:
            self.pipusub.send([])

    def _spatial_all_wgt(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        wgt_useAC = pnwgt.Checkbox(
            name="UseAC", value=self._useAC, width=120, height=15
        )

        def callback_useAC(val):
            self._useAC = val.new
            self.update_AC()

        wgt_useAC.param.watch(callback_useAC, "value")
        return pn.layout.WidgetBox(wgt_useAC, width=150)

    def _spatial_all(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        metas = self.metas
        Asum = regrid(
            hv.Image(self.Asum.sel(**metas), ["width", "height"]), precompute=True
        ).opts(
            plot=dict(frame_height=len(self._h), frame_width=len(self._w)),
            style=dict(cmap="Viridis"),
        )
        cents = (
            hv.Dataset(
                self.cents_sub.drop(list(self.meta_dicts.keys()), axis="columns"),
                kdims=["width", "height", "unit_id"],
            )
            .to(hv.Points, ["width", "height"])
            .opts(
                style=dict(
                    alpha=0.1,
                    line_alpha=0,
                    size=5,
                    nonselection_alpha=0.1,
                    selection_alpha=0.9,
                )
            )
            .collate()
            .overlay("unit_id")
            .opts(plot=dict(tools=["hover", "box_select"]))
        )
        self.strm_uid.source = cents
        fim = fct.partial(hv.Image, kdims=["width", "height"])
        AC = regrid(hv.DynamicMap(fim, streams=[self.pipAC]), precompute=True).opts(
            plot=dict(frame_height=len(self._h), frame_width=len(self._w)),
            style=dict(cmap="Viridis"),
        )
        mov = regrid(hv.DynamicMap(fim, streams=[self.pipmov]), precompute=True).opts(
            plot=dict(frame_height=len(self._h), frame_width=len(self._w)),
            style=dict(cmap="Viridis"),
        )
        lab = fct.partial(hv.Labels, kdims=["width", "height"], vdims=["unit_id"])
        ulab = hv.DynamicMap(lab, streams=[self.pipusub]).opts(
            style=dict(text_color="red")
        )
        return pn.panel(Asum * cents + AC * ulab + mov)

    def update_spatial_all(self):
        """[summary]
        """
        self.spatial_all.objects = self._spatial_all().objects


def write_vid_blk(arr, vpath, options):
    """[summary]

    Args:
        arr ([type]): [description]
        vpath ([type]): [description]
        options ([type]): [description]

    Returns:
        [type]: [description]
    """
    uid = uuid4()
    vname = "{}.mp4".format(uid)
    fpath = os.path.join(vpath, vname)
    arr = np.clip(arr, 0, 255).astype(np.uint8)
    container = av.open(fpath, mode="w")
    stream = container.add_stream("libx264", rate=30)
    stream.width = arr.shape[2]
    stream.height = arr.shape[1]
    stream.pix_fmt = "yuv420p"
    stream.options = options
    for fm in arr:
        fm = cv2.cvtColor(fm, cv2.COLOR_GRAY2RGB)
        fmav = av.VideoFrame.from_ndarray(fm, format="rgb24")
        for p in stream.encode(fmav):
            container.mux(p)
    for p in stream.encode():
        container.mux(p)
    container.close()
    return fpath


def write_video(
    arr, vname=None, vpath=".", options={"crf": "18", "preset": "ultrafast"}
):
    """[summary]

    Args:
        arr ([type]): [description]
        vname ([type], optional): [description]. Defaults to None.
        vpath (str, optional): [description]. Defaults to ".".
        options (dict, optional): [description]. Defaults to {'crf': '18', 'preset': 'ultrafast'}.

    Returns:
        [type]: [description]
    """
    if not vname:
        vname = "{}.mp4".format(uuid4())
    fname = os.path.join(vpath, vname)
    paths = [
        dask.delayed(write_vid_blk)(np.asscalar(a), vpath, options)
        for a in arr.data.to_delayed()
    ]
    with dask.config.set(scheduler="processes"):
        paths = dask.compute(paths)[0]
    streams = [ffmpeg.input(p) for p in paths]
    (ffmpeg.concat(*streams).output(fname).run(overwrite_output=True))
    for vp in paths:
        os.remove(vp)
    return fname


def generate_videos(
    minian,
    varr,
    vpath=".",
    vname="minian.mp4",
    scale="auto",
    options={"crf": "18", "preset": "ultrafast"},
):
    """[summary]

    Args:
        minian ([type]): [description]
        varr ([type]): [description]
        vpath (str, optional): [description]. Defaults to ".".
        vname (str, optional): [description]. Defaults to "minian.mp4".
        scale (str, optional): [description]. Defaults to "auto".
        options (dict, optional): [description]. Defaults to {'crf': '18', 'preset': 'ultrafast'}.

    Returns:
        [type]: [description]
    """
    print("generating traces")
    A = minian["A"].compute().transpose("unit_id", "height", "width")
    C = minian["C"].chunk(dict(unit_id=-1)).transpose("frame", "unit_id")
    Y = (
        minian["Y"]
        .chunk(dict(height=-1, width=-1))
        .transpose("frame", "height", "width")
    )
    org = varr
    try:
        bl = minian["bl"].chunk(dict(unit_id=-1))
    except KeyError:
        print("cannot find background term")
        bl = 0
    C = C + bl
    AC = xr.apply_ufunc(
        da.tensordot,
        C,
        A,
        input_core_dims=[["frame", "unit_id"], ["unit_id", "height", "width"]],
        output_core_dims=[["frame", "height", "width"]],
        dask="allowed",
        kwargs=dict(axes=(1, 0)),
        output_dtypes=[A.dtype],
    )
    org_norm = org
    if scale == "auto":
        Y_max = Y.max().compute()
        Y_norm = Y * (255 / Y_max)
        AC_norm = AC * (255 / Y_max)
    else:
        Y_norm = Y * scale
        AC_norm = AC * scale
    res_norm = Y_norm - AC_norm
    print("writing videos")
    path_org = write_video(org_norm, vpath=vpath, options=options)
    path_Y = write_video(Y_norm, vpath=vpath, options=options)
    path_AC = write_video(AC_norm, vpath=vpath, options=options)
    path_res = write_video(res_norm, vpath=vpath, options=options)
    print("concatenating results")
    str_org = ffmpeg.input(path_org)
    str_Y = ffmpeg.input(path_Y)
    str_AC = ffmpeg.input(path_AC)
    str_res = ffmpeg.input(path_res)
    vtop = ffmpeg.filter([str_org, str_Y], "hstack")
    vbot = ffmpeg.filter([str_res, str_AC], "hstack")
    vid = ffmpeg.filter([vtop, vbot], "vstack")
    fname = os.path.join(vpath, vname)
    vid.output(fname).overwrite_output().run()
    for p in [path_res, path_AC, path_Y, path_org]:
        os.remove(p)
    return fname


def datashade_ndcurve(ovly, kdim=None, spread=False):
    """[summary]

    Args:
        ovly ([type]): [description]
        kdim ([type], optional): [description]. Defaults to None.
        spread (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    if not kdim:
        kdim = ovly.kdims[0].name
    var = np.unique(ovly.dimension_values(kdim)).tolist()
    color_key = [(v, Category10_10[iv]) for iv, v in enumerate(var)]
    color_pts = hv.NdOverlay(
        {
            k: hv.Points([0, 0], label=str(k)).opts(style=dict(color=v))
            for k, v in color_key
        }
    )
    ds_ovly = datashade(
        ovly,
        aggregator=count_cat(kdim),
        color_key=dict(color_key),
        min_alpha=200,
        normalization="linear",
    )
    if spread:
        ds_ovly = dynspread(ds_ovly)
    return ds_ovly * color_pts


def construct_G(g, T):
    """[summary]

    Args:
        g ([type]): [description]
        T ([type]): [description]

    Returns:
        [type]: [description]
    """
    cur_c, cur_r = np.zeros(T), np.zeros(T)
    cur_c[0] = 1
    cur_r[0] = 1
    cur_c[1 : len(g) + 1] = -g
    return linalg.toeplitz(cur_c, cur_r)


def normalize(a):
    """[summary]

    Args:
        a ([type]): [description]

    Returns:
        [type]: [description]
    """
    return np.interp(a, (np.nanmin(a), np.nanmax(a)), (0, +1))


def convolve_G(s, g):
    """[summary]

    Args:
        s ([type]): [description]
        g ([type]): [description]

    Returns:
        [type]: [description]
    """
    G = construct_G(g, len(s))
    try:
        c = np.linalg.inv(G).dot(s)
    except np.linalg.LinAlgError:
        c = s.copy()
    return c


def construct_pulse_response(g, length=500):
    """[summary]

    Args:
        g ([type]): [description]
        length (int, optional): [description]. Defaults to 500.

    Returns:
        [type]: [description]
    """
    s = np.zeros(length)
    s[np.arange(0, length, 500)] = 1
    c = convolve_G(s, g)
    return s, c


def centroid(A, verbose=False):
    """[summary]

    Args:
        A ([type]): [description]
        verbose (bool, optional): [description]. Defaults to False.
    """

    def rel_cent(im):
        """[summary]

        Args:
            im ([type]): [description]

        Returns:
            [type]: [description]
        """
        im_nan = np.isnan(im)
        if im_nan.all():
            return np.array([np.nan, np.nan])
        if im_nan.any():
            im = np.nan_to_num(im)
        cent = np.array(center_of_mass(im))
        return cent / im.shape

    gu_rel_cent = da.gufunc(
        rel_cent,
        signature="(h,w)->(d)",
        output_dtypes=float,
        output_sizes=dict(d=2),
        vectorize=True,
    )
    cents = xr.apply_ufunc(
        gu_rel_cent,
        A.chunk(dict(height=-1, width=-1)),
        input_core_dims=[["height", "width"]],
        output_core_dims=[["dim"]],
        dask="allowed",
    ).assign_coords(dim=["height", "width"])
    if verbose:
        print("computing centroids")
        with ProgressBar():
            cents = cents.compute()
    cents_df = (
        cents.rename("cents")
        .to_series()
        .dropna()
        .unstack("dim")
        .rename_axis(None, axis="columns")
        .reset_index()
    )
    h_rg = (A.coords["height"].min().values, A.coords["height"].max().values)
    w_rg = (A.coords["width"].min().values, A.coords["width"].max().values)
    cents_df["height"] = cents_df["height"] * (h_rg[1] - h_rg[0]) + h_rg[0]
    cents_df["width"] = cents_df["width"] * (w_rg[1] - w_rg[0]) + w_rg[0]
    return cents_df


def visualize_preprocess(fm, fn=None, include_org=True, **kwargs):
    """[summary]

    Args:
        fm ([type]): [description]
        fn ([type], optional): [description]. Defaults to None.
        include_org (bool, optional): [description]. Defaults to True.

    Returns:
        [type]: [description]
    """
    fh, fw = fm.sizes["height"], fm.sizes["width"]
    asp = fw / fh
    opts_im = {
        "plot": {
            "frame_width": 500,
            "aspect": asp,
            "title": "Image {label} {group} {dimensions}",
        },
        "style": {"cmap": "viridis"},
    }
    opts_cnt = {
        "plot": {
            "frame_width": 500,
            "aspect": asp,
            "title": "Contours {label} {group} {dimensions}",
        },
        "style": {"cmap": "viridis"},
    }

    def _vis(f):
        """[summary]

        Args:
            f ([type]): [description]

        Returns:
            [type]: [description]
        """
        im = hv.Image(f, kdims=["width", "height"]).opts(**opts_im)
        cnt = hv.operation.contours(im).opts(**opts_cnt)
        return im, cnt

    if fn is not None:
        pkey = kwargs.keys()
        pval = kwargs.values()
        im_dict = dict()
        cnt_dict = dict()
        for params in itt.product(*pval):
            fm_res = fn(fm, **dict(zip(pkey, params)))
            cur_im, cur_cnt = _vis(fm_res)
            cur_im = cur_im.relabel("After")
            cur_cnt = cur_cnt.relabel("After")
            p_str = tuple(
                [str(p) if not isinstance(p, (int, float)) else p for p in params]
            )
            im_dict[p_str] = cur_im
            cnt_dict[p_str] = cur_cnt
        hv_im = regrid(hv.HoloMap(im_dict, kdims=list(pkey)), precompute=True).opts(
            **opts_im
        )
        hv_cnt = datashade(
            hv.HoloMap(cnt_dict, kdims=list(pkey)), precompute=True, cmap=Viridis256
        ).opts(**opts_cnt)
        if include_org:
            im, cnt = _vis(fm)
            im = regrid(im, precompute=True).relabel("Before").opts(**opts_im)
            cnt = (
                datashade(cnt, precompute=True, cmap=Viridis256)
                .relabel("Before")
                .opts(**opts_cnt)
            )
        return (im + cnt + hv_im + hv_cnt).cols(2)
    else:
        im, cnt = _vis(fm)
        im = im.relabel("Before")
        cnt = cnt.relabel("Before")
        return im + cnt


def visualize_seeds(max_proj, seeds, mask=None, datashade=False):
    """[summary]

    Args:
        max_proj ([type]): [description]
        seeds ([type]): [description]
        mask ([type], optional): [description]. Defaults to None.
        datashade (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    h, w = max_proj.sizes["height"], max_proj.sizes["width"]
    asp = w / h
    pt_cmap = {True: "white", False: "red"}
    opts_im = dict(plot=dict(frame_width=600, aspect=asp), style=dict(cmap="Viridis"))
    opts_pts = dict(
        plot=dict(
            frame_width=600,
            aspect=asp,
            size_index="seeds",
            color_index=mask,
            tools=["hover"],
        ),
        style=dict(fill_alpha=0.8, line_alpha=0, cmap=pt_cmap),
    )
    if mask:
        vdims = ["index", "seeds", mask]
    else:
        vdims = ["index", "seeds"]
        opts_pts["style"]["color"] = "white"
    im = hv.Image(max_proj, kdims=["width", "height"])
    pts = hv.Points(seeds, kdims=["width", "height"], vdims=vdims)
    if datashade:
        im = regrid(im)
    return im.opts(**opts_im) * pts.opts(**opts_pts)


def visualize_gmm_fit(values, gmm, bins):
    """[summary]

    Args:
        values ([type]): [description]
        gmm ([type]): [description]
        bins ([type]): [description]
    """

    def gaussian(x, mu, sig):
        return np.exp(-np.power(x - mu, 2.0) / (2 * np.power(sig, 2.0)))

    hist = np.histogram(values, bins=bins, density=True)
    gss_dict = dict()
    for igss, (mu, sig) in enumerate(zip(gmm.means_, gmm.covariances_)):
        gss = gaussian(hist[1], np.asscalar(mu), np.asscalar(np.sqrt(sig)))
        gss_dict[igss] = hv.Curve((hist[1], gss))
    return (
        hv.Histogram(((hist[0] - hist[0].min()) / np.ptp(hist[0]), hist[1])).opts(
            style=dict(alpha=0.6, fill_color="gray")
        )
        * hv.NdOverlay(gss_dict)
    ).opts(plot=dict(height=350, width=500))


def visualize_spatial_update(A_dict, C_dict, kdims=None, norm=True, datashading=True):
    """[summary]

    Args:
        A_dict ([type]): [description]
        C_dict ([type]): [description]
        kdims ([type], optional): [description]. Defaults to None.
        norm (bool, optional): [description]. Defaults to True.
        datashading (bool, optional): [description]. Defaults to True.

    Returns:
        [type]: [description]
    """
    if not kdims:
        A_dict = dict(dummy=A_dict)
        C_dict = dict(dummy=C_dict)
    hv_pts_dict, hv_A_dict, hv_Ab_dict, hv_C_dict = (dict(), dict(), dict(), dict())
    for key, A in A_dict.items():
        A = A.compute()
        C = C_dict[key]
        if norm:
            C = xr.apply_ufunc(
                normalize,
                C.chunk(dict(frame=-1)),
                input_core_dims=[["frame"]],
                output_core_dims=[["frame"]],
                vectorize=True,
                dask="parallelized",
                output_dtypes=[C.dtype],
            )
        C = C.compute()
        h, w = A.sizes["height"], A.sizes["width"]
        cents_df = centroid(A)
        hv_pts_dict[key] = hv.Points(
            cents_df, kdims=["width", "height"], vdims=["unit_id"]
        ).opts(
            plot=dict(tools=["hover"]), style=dict(fill_alpha=0.2, line_alpha=0, size=8)
        )
        hv_A_dict[key] = hv.Image(
            A.sum("unit_id").rename("A"), kdims=["width", "height"]
        )
        hv_Ab_dict[key] = hv.Image(
            (A > 0).sum("unit_id").rename("A_bin"), kdims=["width", "height"]
        )
        hv_C_dict[key] = hv.Dataset(C.rename("C")).to(hv.Curve, kdims="frame")
    hv_pts = hv.HoloMap(hv_pts_dict, kdims=kdims)
    hv_A = hv.HoloMap(hv_A_dict, kdims=kdims)
    hv_Ab = hv.HoloMap(hv_Ab_dict, kdims=kdims)
    hv_C = (
        hv.HoloMap(hv_C_dict, kdims=kdims)
        .collate()
        .grid("unit_id")
        .add_dimension("time", 0, 0)
    )
    if datashading:
        hv_A = regrid(hv_A)
        hv_Ab = regrid(hv_Ab)
        hv_C = datashade(hv_C)
    hv_A = hv_A.opts(frame_width=400, aspect=w / h, colorbar=True, cmap="viridis")
    hv_Ab = hv_Ab.opts(frame_width=400, aspect=w / h, colorbar=True, cmap="viridis")
    hv_C = hv_C.map(
        lambda cr: cr.opts(frame_width=500, frame_height=50),
        hv.RGB if datashading else hv.Curve,
    )
    return hv.NdLayout(
        {"pseudo-color": (hv_pts * hv_A), "binary": (hv_pts * hv_Ab)},
        kdims="Spatial Matrix",
    ).cols(1) + hv_C.relabel("Temporal Components")


def visualize_temporal_update(
    YA_dict,
    C_dict,
    S_dict,
    g_dict,
    sig_dict,
    A_dict,
    kdims=None,
    norm=True,
    datashading=True,
):
    """[summary]

    Args:
        YA_dict ([type]): [description]
        C_dict ([type]): [description]
        S_dict ([type]): [description]
        g_dict ([type]): [description]
        sig_dict ([type]): [description]
        A_dict ([type]): [description]
        kdims ([type], optional): [description]. Defaults to None.
        norm (bool, optional): [description]. Defaults to True.
        datashading (bool, optional): [description]. Defaults to True.

    Returns:
        [type]: [description]
    """
    inputs = [YA_dict, C_dict, S_dict, sig_dict, g_dict]
    if not kdims:
        inputs = [dict(dummy=i) for i in inputs]
        A_dict = dict(dummy=A_dict)
    input_dict = {k: [i[k] for i in inputs] for k in inputs[0].keys()}
    hv_YA, hv_C, hv_S, hv_sig, hv_C_pul, hv_S_pul, hv_A = [dict() for _ in range(7)]
    for k, ins in input_dict.items():
        if norm:
            ins[:-1] = [
                xr.apply_ufunc(
                    normalize,
                    i.chunk(dict(frame=-1)),
                    input_core_dims=[["frame"]],
                    output_core_dims=[["frame"]],
                    vectorize=True,
                    dask="parallelized",
                    output_dtypes=[i.dtype],
                )
                for i in ins[:-1]
            ]
        ins[:] = [i.compute() for i in ins]
        ya, c, s, sig, g = ins
        f_crd = ya.coords["frame"]
        pul_crd = f_crd.values[:500]
        s_pul, c_pul = xr.apply_ufunc(
            construct_pulse_response,
            g,
            input_core_dims=[["lag"]],
            output_core_dims=[["t"], ["t"]],
            vectorize=True,
            kwargs=dict(length=len(pul_crd)),
            output_sizes=dict(t=len(pul_crd)),
        )
        s_pul, c_pul = (s_pul.assign_coords(t=pul_crd), c_pul.assign_coords(t=pul_crd))
        if norm:
            c_pul = xr.apply_ufunc(
                normalize,
                c_pul.chunk(dict(t=-1)),
                input_core_dims=[["t"]],
                output_core_dims=[["t"]],
                dask="parallelized",
                output_dtypes=[c_pul.dtype],
            ).compute()
        pul_range = (
            f_crd.min(),
            int(np.around(f_crd.min() + (f_crd.max() - f_crd.min()) / 2)),
        )
        hv_S_pul[k], hv_C_pul[k] = [
            (hv.Dataset(tr.rename("Response (A.U.)")).to(hv.Curve, kdims=["t"]))
            for tr in [s_pul, c_pul]
        ]
        hv_YA[k] = hv.Dataset(ya.rename("Intensity (A.U.)")).to(
            hv.Curve, kdims=["frame"]
        )
        if c.sizes["unit_id"] > 0:
            hv_C[k], hv_S[k], hv_sig[k] = [
                (
                    hv.Dataset(tr.rename("Intensity (A.U.)")).to(
                        hv.Curve, kdims=["frame"]
                    )
                )
                for tr in [c, s, sig]
            ]
        hv_A[k] = hv.Dataset(A_dict[k].rename("A")).to(
            hv.Image, kdims=["width", "height"]
        )
        h, w = A_dict[k].sizes["height"], A_dict[k].sizes["width"]
    hvobjs = [hv_YA, hv_C, hv_S, hv_sig, hv_C_pul, hv_S_pul, hv_A]
    hvobjs[:] = [hv.HoloMap(hvobj, kdims=kdims).collate() for hvobj in hvobjs]
    hv_unit = {
        "Raw Signal": hvobjs[0],
        "Fitted Calcium Trace": hvobjs[1],
        "Fitted Spikes": hvobjs[2],
        "Fitted Signal": hvobjs[3],
    }
    hv_pul = {"Simulated Calcium": hvobjs[4], "Simulated Spike": hvobjs[5]}
    hv_unit = hv.HoloMap(hv_unit, kdims="traces").collate().overlay("traces")
    hv_pul = hv.HoloMap(hv_pul, kdims="traces").collate().overlay("traces")
    hv_A = hvobjs[6]
    if datashading:
        hv_unit = datashade_ndcurve(hv_unit, "traces")
        hv_A = regrid(hv_A)
    else:
        hv_unit = Dynamic(hv_unit)
        hv_A = Dynamic(hv_A)
    hv_pul = Dynamic(hv_pul)
    hv_unit = hv_unit.map(
        lambda p: p.opts(plot=dict(frame_height=400, frame_width=1000))
    )
    hv_pul = hv_pul.opts(plot=dict(frame_width=500, aspect=w / h)).redim(
        t=hv.Dimension("t", soft_range=pul_range)
    )
    hv_A = hv_A.opts(
        plot=dict(frame_width=500, aspect=w / h), style=dict(cmap="Viridis")
    )
    return (
        hv_unit.relabel("Current Unit: Temporal Traces")
        + hv.NdLayout(
            {"Simulated Pulse Response": hv_pul, "Spatial Footprint": hv_A},
            kdims="Current Unit",
        )
    ).cols(1)


def NNsort(cents):
    """[summary]

    Args:
        cents ([type]): [description]

    Returns:
        [type]: [description]
    """
    cents_hw = cents[["height", "width"]]
    kdtree = cKDTree(cents_hw)
    idu_start = cents_hw.sum(axis="columns").idxmin()
    result = pd.Series(0, index=cents.index)
    remain_list = cents.index.tolist()
    idu_next = idu_start
    NNord = 0
    while remain_list:
        result.loc[idu_next] = NNord
        remain_list.remove(idu_next)
        for k in range(1, int(np.ceil(np.log2(len(result)))) + 1):
            qry = kdtree.query(cents_hw.loc[idu_next], 2 ** k)
            NNs = qry[1][np.isfinite(qry[0])].squeeze()
            NNs = NNs[np.sort(np.unique(NNs, return_index=True)[1])]
            NNs = np.array(result.iloc[NNs].index)
            NN_idxs = np.argwhere(np.isin(NNs, remain_list, assume_unique=True))
            if len(NN_idxs) > 0:
                NN = NNs[NN_idxs[0]][0]
                idu_next = NN
                NNord = NNord + 1
                break
    return result



import argparse
import os
import requests

PIPELINE_FILES = ("pipeline.ipynb","pipeline_noted.ipynb","cross-registration.ipynb")
DEMO_FILES = [f"demo_movies/msCam{i}.avi" for i in range(1,11)]

def _get_file(filename: str, branch: str):
  if os.path.isfile(f"{filename}"):
    print(f"File {filename} already exists, skipping install of this file.")
    return

  r = requests.get(f"https://raw.github.com/DeniseCaiLab/minian/{branch}/{filename}")
  if r.status_code == 200:
    with open(f"{filename}","wb") as f:
      for chunk in r.iter_content(2048):
        f.write(chunk)
    print(f"File {filename} installed.")

def demo(branch: str):
  try:
    os.mkdir("demo_movies")
    print("Installing demo movies")
    for file in DEMO_FILES:
      _get_file(file, branch)
  except OSError:
    print("Creation of the directory demo_movies failed, not installing.")

def pipeline(branch: str):
  print("Installing pipeline notebooks")
  for file in PIPELINE_FILES:
    _get_file(file, branch)

def main():
        parser = argparse.ArgumentParser()
        parser.add_argument("--notebooks", action="store_true", help="Installs the pipeline notebooks")
        parser.add_argument("--demo", action="store_true", help="Installs the demo movies")
        parser.add_argument("-b", action="store", default="master", help="Git repo branch name, default master")
        args = parser.parse_args()

        branch = args.b
        print(f"Using branch: {branch}")

        if args.notebooks:
          pipeline(branch)

        if args.demo:
          demo(branch)



from minian.utilities import save_minian
import numpy as np
from numpy.core.fromnumeric import amin
import xarray as xr
import os
from numpy import random
from ..preprocessing import *
from ..motion_correction import *
from ..initialization import *
from ..cnmf import *
from ..visualization import convolve_G, write_video
from cv2 import GaussianBlur


def gauss_cell(
    height: int, width: int, sigma: float, cov_coef: float, cent=None, nsamp=1000
):
    # generate centroid
    if cent is None:
        cent = (random.randint(height), random.randint(width))
    # generate covariance
    while True:
        cov_var = random.rand(2, 2)
        cov_var = (cov_var + cov_var.T) / 2 * cov_coef
        cov = np.eye(2) * sigma + cov_var
        if np.all(np.linalg.eigvals(cov) > 0):
            break  # ensure cov is positive definite
    # generate samples of coordinates
    crds = np.clip(
        np.round(random.multivariate_normal(cent, cov, size=nsamp)).astype(np.int),
        0,
        None,
    )
    # generate spatial footprint
    A = np.zeros((height, width))
    for crd in np.unique(crds, axis=0):
        try:
            A[tuple(crd)] = np.sum(np.all(crds == crd, axis=1))
        except IndexError:
            pass
    return A / A.max()


def ar_trace(frame: int, pfire: float, g: np.ndarray):
    S = random.binomial(n=1, p=pfire, size=frame)
    C = convolve_G(S, g)
    return C, S


def generate_data(
    ncell: int,
    dims: dict,
    sp_noise: float,
    tmp_noise: float,
    sp_sigma: float,
    sp_cov_coef: float,
    tmp_pfire: float,
    tmp_g_avg: float,
    tmp_g_var: float,
    bg_sigma=0,
    bg_strength=0,
    mo_sigma=0,
):
    ff, hh, ww = (
        dims["frame"],
        dims["height"],
        dims["width"],
    )
    hh_pad, ww_pad = hh + mo_sigma * 4, ww + mo_sigma * 4
    A = xr.DataArray(
        np.stack(
            [
                gauss_cell(hh_pad, ww_pad, sp_sigma, cov_coef=sp_cov_coef)
                for _ in range(ncell)
            ]
        ),
        dims=["unit_id", "height", "width"],
        coords={
            "height": np.arange(hh_pad),
            "width": np.arange(ww_pad),
            "unit_id": np.arange(ncell),
        },
        name="A",
    )
    tmp_g = np.clip(
        random.normal(tmp_g_avg, tmp_g_var, size=ncell), a_min=0.8, a_max=0.95
    )
    traces = [ar_trace(ff, tmp_pfire, np.array([g])) for g in tmp_g]
    C = xr.DataArray(
        np.stack([t[0] for t in traces]),
        dims=["unit_id", "frame"],
        coords={"unit_id": np.arange(ncell), "frame": np.arange(ff)},
        name="C",
    )
    S = xr.DataArray(
        np.stack([t[1] for t in traces]),
        dims=["unit_id", "frame"],
        coords={"unit_id": np.arange(ncell), "frame": np.arange(ff)},
        name="S",
    )
    C_noise = C + random.normal(scale=tmp_noise, size=(ncell, ff))
    Y = C_noise.dot(A).rename("Y")
    Y = Y / Y.max()
    if bg_strength:
        A_bg = xr.apply_ufunc(
            GaussianBlur,
            A,
            input_core_dims=[["height", "width"]],
            output_core_dims=[["height", "width"]],
            vectorize=True,
            kwargs={
                "ksize": (int(hh // 2 * 2 - 1), int(ww // 2 * 2 - 1)),
                "sigmaX": bg_sigma,
            },
        )
        Y_bg = C_noise.dot(A_bg)
        Y_bg = Y_bg / Y_bg.max()
        Y = Y + Y_bg * bg_strength
    shifts = xr.DataArray(
        np.clip(
            random.normal(scale=mo_sigma, size=(ff, 2)),
            a_min=-2 * mo_sigma,
            a_max=2 * mo_sigma,
        ),
        dims=["frame", "variable"],
        coords={"frame": np.arange(ff), "variable": ["height", "width"]},
        name="shifts",
    )
    Y = (
        apply_shifts(Y, shifts)
        .compute()
        .isel(
            height=slice(2 * mo_sigma, -2 * mo_sigma),
            width=slice(2 * mo_sigma, -2 * mo_sigma),
        )
    )
    Y = (Y / Y.max() + random.normal(scale=sp_noise, size=(ff, hh, ww))).rename("Y")
    return Y, A, C, S, shifts


if __name__ == "__main__":
    # generate toy data
    testpath = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data", "toy")
    Y, A, C, S, shifts = generate_data(
        ncell=100,
        dims={"height": 100, "width": 100, "frame": 1000},
        sp_noise=0.05,
        tmp_noise=0.08,
        sp_sigma=3,
        sp_cov_coef=2,
        tmp_pfire=0.1,
        tmp_g_avg=0.9,
        tmp_g_var=0.03,
        bg_sigma=20,
        bg_strength=1,
        mo_sigma=1,
    )
    Y_true = C.dot(A)
    Y_true = (
        ((Y_true - Y_true.min()) / (Y_true.max() - Y_true.min()) * 255)
        .astype(np.uint8)
        .rename("Y_true")
    )
    Y = (
        ((Y - Y.min()) / (Y.max() - Y.min()) * 255)
        .astype(np.uint8)
        .chunk({"frame": 500})
    )
    write_video(Y, "toy.mp4", testpath)
    # run pipeline
    print("pre-processing")
    Y_glow = (Y - Y.min("frame").compute()).rename("Y_glow")
    Y_dn = denoise(Y_glow, method="median", ksize=5).rename("Y_denoise")
    Y_bg = remove_background(Y_dn, method="tophat", wnd=5).rename("Y_bg")
    shifts_est = estimate_shifts(Y_bg, max_sh=10)
    Y_mc = apply_shifts(Y_bg, shifts_est).fillna(0).rename("Y_mc")
    print("initialization")
    Y_flt = Y.compute().stack(spatial=["height", "width"])
    seeds_in = seeds_init(Y_mc, wnd_size=200, stp_size=100, diff_thres=2)
    seeds_pnr, _, _ = pnr_refine(Y_flt, seeds_in.copy(), noise_freq=0.01, thres=0.5)
    seeds_ks = ks_refine(Y_flt, seeds_pnr[seeds_pnr["mask_pnr"]])
    seeds_mrg = seeds_merge(
        Y_flt,
        seeds_ks[seeds_ks["mask_ks"]].reset_index(drop=True),
        thres_dist=2,
        thres_corr=0.7,
        noise_freq=0.01,
    )
    A_init, C_init, b_init, f_init = initialize(
        Y_mc, seeds_mrg[seeds_mrg["mask_mrg"]], thres_corr=0.8, wnd=15, noise_freq=0.01
    )
    print("cnmf")
    sn = get_noise_fft(Y_mc).persist()
    A_sp1, b_sp1, C_sp1, f_sp1 = update_spatial(
        Y_mc, A_init, b_init, C_init, f_init, sn
    )
    YrA, C_tp1, S_tp1, B_tp1, C0_tp1, sig_tp1, g_tp1, scale_tp1 = update_temporal(
        Y_mc, A_sp1, b_sp1, C_sp1, f_sp1, sn
    )
    A_tp1 = A_sp1.sel(unit_id=C_tp1.coords["unit_id"]).rename("A_tp1")
    A_mrg, sig_mrg = unit_merge(A_tp1, sig_tp1, [])
    print(A_mrg)
    # save results
    (
        A_init,
        C_init,
        b_init,
        f_init,
        sn,
        A_sp1,
        b_sp1,
        C_sp1,
        f_sp1,
        C_tp1,
        S_tp1,
        sig_tp1,
        A_mrg,
        sig_mrg,
    ) = (
        A_init.rename("A_init"),
        C_init.rename("C_init"),
        b_init.rename("b_init"),
        f_init.rename("f_init"),
        sn.rename("sn"),
        A_sp1.rename("A_sp1"),
        b_sp1.rename("b_sp1"),
        C_sp1.rename("C_sp1"),
        f_sp1.rename("f_sp1"),
        C_tp1.rename("C_tp1"),
        S_tp1.rename("S_tp1"),
        sig_tp1.rename("sig_tp1"),
        A_mrg.rename("A_mrg"),
        sig_mrg.rename("sig_mrg"),
    )
    for dat in [
        Y,
        A,
        C,
        S,
        shifts,
        Y_true,
        Y_glow,
        Y_dn,
        Y_bg,
        Y_mc,
        A_init,
        C_init,
        b_init,
        f_init,
        sn,
        A_sp1,
        b_sp1,
        C_sp1,
        f_sp1,
        C_tp1,
        S_tp1,
        sig_tp1,
        A_mrg,
        sig_mrg,
    ]:
        save_minian(
            dat,
            dpath=testpath,
            fname="minian",
            backend="zarr",
            meta_dict={"session": -1, "animal": -2},
            overwrite=True,
        )




import cv2
import xarray as xr
from scipy.ndimage import uniform_filter
from skimage.morphology import disk


def remove_background(varr, method, wnd):
    """[summary]

    Args:
        varr ([type]): [description]
        method ([type]): [description]
        wnd ([type]): [description]

    Returns:
        [type]: [description]
    """
    selem = disk(wnd)
    res = xr.apply_ufunc(
        remove_background_perframe,
        varr.chunk(dict(height=-1, width=-1)),
        input_core_dims=[["height", "width"]],
        output_core_dims=[["height", "width"]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[varr.dtype],
        kwargs=dict(method=method, wnd=wnd, selem=selem),
    )
    res = res.astype(varr.dtype)
    return res.rename(varr.name + "_subtracted")


def remove_background_perframe(fm, method, wnd, selem):
    """[summary]

    Args:
        fm ([type]): [description]
        method ([type]): [description]
        wnd ([type]): [description]
        selem ([type]): [description]

    Returns:
        [type]: [description]
    """
    if method == "uniform":
        return fm - uniform_filter(fm, wnd)
    elif method == "tophat":
        return cv2.morphologyEx(fm, cv2.MORPH_TOPHAT, selem)


def stripe_correction(varr, reduce_dim="height", on="mean"):
    """[summary]

    Args:
        varr ([type]): [description]
        reduce_dim (str, optional): [description]. Defaults to 'height'.
        on (str, optional): [description]. Defaults to 'mean'.

    Raises:
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """
    if on == "mean":
        temp = varr.mean(dim="frame")
    elif on == "max":
        temp = varr.max(dim="frame")
    elif on == "perframe":
        temp = varr
    else:
        raise NotImplementedError("on {} not understood".format(on))
    mean1d = temp.mean(dim=reduce_dim)
    varr_sc = varr - mean1d
    return varr_sc.rename(varr.name + "_Stripe_Corrected")


def gaussian_blur(varray, ksize=(3, 3), sigmaX=0):
    """[summary]

    Args:
        varray ([type]): [description]
        ksize (tuple, optional): [description]. Defaults to (3, 3).
        sigmaX (int, optional): [description]. Defaults to 0.

    Returns:
        [type]: [description]
    """
    return varray.groupby("frame").apply(
        lambda fm: cv2.GaussianBlur(fm.values, ksize, sigmaX)
    )


def denoise(varr, method, **kwargs):
    """[summary]

    Args:
        varr ([type]): [description]
        method ([type]): [description]

    Raises:
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """
    if method == "gaussian":
        func = cv2.GaussianBlur
    elif method == "median":
        func = cv2.medianBlur
    elif method == "bilateral":
        func = cv2.bilateralFilter
    else:
        raise NotImplementedError("denoise method {} not understood".format(method))
    res = xr.apply_ufunc(
        func,
        varr,
        input_core_dims=[["height", "width"]],
        output_core_dims=[["height", "width"]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[varr.dtype],
        kwargs=kwargs,
    )
    res = res.astype(varr.dtype)
    return res.rename(varr.name + "_denoised")
